#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


service_chain_opts = [
    cfg.ListOpt('servicechain_drivers',
                default=['dummy'],
                help=_("An ordered list of service chain drivers "
                       "entrypoints to be loaded from the "
                       "gbpservice.neutron.servicechain.servicechain_drivers "
                       "namespace."))
]


cfg.CONF.register_opts(service_chain_opts, "servicechain")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron._i18n import _LE
from neutron.plugins.common import constants as pconst
from oslo_log import helpers as log
from oslo_log import log as logging
from oslo_utils import excutils

import gbpservice.neutron.db.servicechain_db as servicechain_db
from gbpservice.neutron.services.grouppolicy.common import constants as gp_cts
from gbpservice.neutron.services.servicechain.plugins.msc import (
    context as servicechain_context)
from gbpservice.neutron.services.servicechain.plugins.msc import (
    driver_manager as manager)
from gbpservice.neutron.services.servicechain.plugins import sharing


LOG = logging.getLogger(__name__)


class ServiceChainPlugin(servicechain_db.ServiceChainDbPlugin,
                         sharing.SharingMixin):

    """Implementation of the Service Chain Plugin.

    """
    supported_extension_aliases = ["servicechain"]
    path_prefix = gp_cts.GBP_PREFIXES[pconst.SERVICECHAIN]

    def __init__(self):
        self.driver_manager = manager.DriverManager()
        super(ServiceChainPlugin, self).__init__()
        self.driver_manager.initialize()

    @log.log_method_call
    def create_servicechain_node(self, context, servicechain_node):
        session = context.session
        with session.begin(subtransactions=True):
            result = super(ServiceChainPlugin, self).create_servicechain_node(
                context, servicechain_node)
            self._validate_shared_create(context, result, 'servicechain_node')
            sc_context = servicechain_context.ServiceChainNodeContext(
                self, context, result)
            self.driver_manager.create_servicechain_node_precommit(
                sc_context)

        try:
            self.driver_manager.create_servicechain_node_postcommit(
                sc_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("driver_manager.create_servicechain_postcommit "
                              "failed, deleting servicechain_node %s"),
                          result['id'])
                self.delete_servicechain_node(context, result['id'])

        return result

    @log.log_method_call
    def update_servicechain_node(self, context, servicechain_node_id,
                                 servicechain_node):
        session = context.session
        with session.begin(subtransactions=True):
            original_sc_node = self.get_servicechain_node(
                                         context, servicechain_node_id)
            updated_sc_node = super(ServiceChainPlugin,
                                    self).update_servicechain_node(
                                        context, servicechain_node_id,
                                        servicechain_node,
                                        set_params=True)
            self._validate_shared_update(context, original_sc_node,
                                         updated_sc_node, 'servicechain_node')
            sc_context = servicechain_context.ServiceChainNodeContext(
                self, context, updated_sc_node,
                original_sc_node=original_sc_node)
            self.driver_manager.update_servicechain_node_precommit(
                sc_context)

        self.driver_manager.update_servicechain_node_postcommit(sc_context)

        return updated_sc_node

    @log.log_method_call
    def delete_servicechain_node(self, context, servicechain_node_id):
        session = context.session
        with session.begin(subtransactions=True):
            sc_node = self.get_servicechain_node(context,
                                                 servicechain_node_id)
            sc_context = servicechain_context.ServiceChainNodeContext(
                self, context, sc_node)
            self.driver_manager.delete_servicechain_node_precommit(
                sc_context)
            super(ServiceChainPlugin, self).delete_servicechain_node(
                context, servicechain_node_id)

        try:
            self.driver_manager.delete_servicechain_node_postcommit(
                sc_context)
        except Exception:
            LOG.exception(_LE("delete_servicechain_node_postcommit failed "
                            "for servicechain_node %s"),
                          servicechain_node_id)

    @log.log_method_call
    def create_servicechain_spec(self, context, servicechain_spec):
        session = context.session
        with session.begin(subtransactions=True):
            result = super(ServiceChainPlugin, self).create_servicechain_spec(
                context, servicechain_spec)
            self._validate_shared_create(context, result, 'servicechain_spec')
            sc_context = servicechain_context.ServiceChainSpecContext(
                self, context, result)
            self.driver_manager.create_servicechain_spec_precommit(
                sc_context)

        try:
            self.driver_manager.create_servicechain_spec_postcommit(sc_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("driver_manager.create_servicechain_postcommit "
                              "failed, deleting servicechain_spec %s"),
                          result['id'])
                self.delete_servicechain_spec(context, result['id'])

        return result

    @log.log_method_call
    def update_servicechain_spec(self, context, servicechain_spec_id,
                                 servicechain_spec):
        session = context.session
        with session.begin(subtransactions=True):
            original_sc_spec = self.get_servicechain_spec(
                                         context, servicechain_spec_id)
            updated_sc_spec = super(ServiceChainPlugin,
                                    self).update_servicechain_spec(
                                        context, servicechain_spec_id,
                                        servicechain_spec)
            self._validate_shared_update(context, original_sc_spec,
                                         updated_sc_spec, 'servicechain_spec')
            sc_context = servicechain_context.ServiceChainSpecContext(
                self, context, updated_sc_spec,
                original_sc_spec=original_sc_spec)
            self.driver_manager.update_servicechain_spec_precommit(
                sc_context)

        self.driver_manager.update_servicechain_spec_postcommit(sc_context)

        return updated_sc_spec

    @log.log_method_call
    def delete_servicechain_spec(self, context, servicechain_spec_id):
        session = context.session
        with session.begin(subtransactions=True):
            sc_spec = self.get_servicechain_spec(context,
                                                 servicechain_spec_id)
            sc_context = servicechain_context.ServiceChainSpecContext(
                self, context, sc_spec)
            self.driver_manager.delete_servicechain_spec_precommit(
                sc_context)
            super(ServiceChainPlugin, self).delete_servicechain_spec(
                context, servicechain_spec_id)

        try:
            self.driver_manager.delete_servicechain_spec_postcommit(sc_context)
        except Exception:
            LOG.exception(_LE("delete_servicechain_spec_postcommit failed "
                            "for servicechain_spec %s"),
                          servicechain_spec_id)

    @log.log_method_call
    def create_servicechain_instance(self, context, servicechain_instance):
        session = context.session
        with session.begin(subtransactions=True):
            result = super(ServiceChainPlugin,
                           self).create_servicechain_instance(
                               context, servicechain_instance)
            sc_context = servicechain_context.ServiceChainInstanceContext(
                self, context, result)
            self.driver_manager.create_servicechain_instance_precommit(
                sc_context)

        try:
            self.driver_manager.create_servicechain_instance_postcommit(
                sc_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE(
                    "driver_manager.create_servicechain_instance_postcommit "
                    "failed, deleting servicechain_instance %s"),
                          result['id'])
                self.delete_servicechain_instance(context, result['id'])

        return result

    @log.log_method_call
    def update_servicechain_instance(self, context,
                                     servicechain_instance_id,
                                     servicechain_instance):
        session = context.session
        with session.begin(subtransactions=True):
            original_sc_instance = self.get_servicechain_instance(
                                         context, servicechain_instance_id)
            updated_sc_instance = super(ServiceChainPlugin,
                                    self).update_servicechain_instance(
                                        context, servicechain_instance_id,
                                        servicechain_instance)
            sc_context = servicechain_context.ServiceChainInstanceContext(
                self, context, updated_sc_instance,
                original_sc_instance=original_sc_instance)
            self.driver_manager.update_servicechain_instance_precommit(
                sc_context)

        self.driver_manager.update_servicechain_instance_postcommit(
            sc_context)
        return updated_sc_instance

    @log.log_method_call
    def delete_servicechain_instance(self, context, servicechain_instance_id):
        session = context.session
        with session.begin(subtransactions=True):
            sc_instance = self.get_servicechain_instance(
                                                context,
                                                servicechain_instance_id)
            sc_context = servicechain_context.ServiceChainInstanceContext(
                self, context, sc_instance)
            self.driver_manager.delete_servicechain_instance_precommit(
                sc_context)
            super(ServiceChainPlugin, self).delete_servicechain_instance(
                context, servicechain_instance_id)

        try:
            self.driver_manager.delete_servicechain_instance_postcommit(
                sc_context)
        except Exception:
            LOG.exception(_LE("delete_servicechain_instance_postcommit failed "
                            "for servicechain_instance %s"),
                          servicechain_instance_id)

    @log.log_method_call
    def create_service_profile(self, context, service_profile):
        session = context.session
        with session.begin(subtransactions=True):
            result = super(ServiceChainPlugin,
                           self).create_service_profile(
                               context, service_profile)
            self._validate_shared_create(context, result, 'service_profile')
            sc_context = servicechain_context.ServiceProfileContext(
                self, context, result)
            self.driver_manager.create_service_profile_precommit(
                sc_context)

        try:
            self.driver_manager.create_service_profile_postcommit(
                sc_context)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE(
                    "driver_manager.create_service_profile_postcommit "
                    "failed, deleting service_profile %s"), result['id'])
                self.delete_service_profile(context, result['id'])

        return result

    @log.log_method_call
    def update_service_profile(self, context, service_profile_id,
                               service_profile):
        session = context.session
        with session.begin(subtransactions=True):
            original_profile = self.get_service_profile(
                context, service_profile_id)
            updated_profile = super(ServiceChainPlugin,
                                    self).update_service_profile(
                context, service_profile_id, service_profile)
            self._validate_shared_update(context, original_profile,
                                         updated_profile, 'service_profile')
            sc_context = servicechain_context.ServiceProfileContext(
                self, context, updated_profile,
                original_profile=original_profile)
            self.driver_manager.update_service_profile_precommit(
                sc_context)

        self.driver_manager.update_service_profile_postcommit(
            sc_context)
        return updated_profile

    @log.log_method_call
    def delete_service_profile(self, context, service_profile_id):
        session = context.session
        with session.begin(subtransactions=True):
            profile = self.get_service_profile(
                context, service_profile_id)
            sc_context = servicechain_context.ServiceProfileContext(
                self, context, profile)
            self.driver_manager.delete_service_profile_precommit(
                sc_context)
            super(ServiceChainPlugin, self).delete_service_profile(
                context, service_profile_id)

        try:
            self.driver_manager.delete_service_profile_postcommit(
                sc_context)
        except Exception:
            LOG.exception(_LE("delete_service_profile_postcommit failed "
                            "for service_profile %s"),
                          service_profile_id)



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.plugins.ml2 import driver_context as ml2_context

from gbpservice.neutron.plugins.ml2plus import driver_api as api


class SubnetPoolContext(ml2_context.MechanismDriverContext,
                        api.SubnetPoolContext):

    def __init__(self, plugin, plugin_context, subnetpool,
                 original_subnetpool=None):
        super(SubnetPoolContext, self).__init__(plugin, plugin_context)
        self._subnetpool = subnetpool
        self._original_subnetpool = original_subnetpool

    @property
    def current(self):
        return self._subnetpool

    @property
    def original(self):
        return self._original_subnetpool


class AddressScopeContext(ml2_context.MechanismDriverContext,
                          api.AddressScopeContext):

    def __init__(self, plugin, plugin_context, address_scope,
                 original_address_scope=None):
        super(AddressScopeContext, self).__init__(plugin, plugin_context)
        self._address_scope = address_scope
        self._original_address_scope = original_address_scope

    @property
    def current(self):
        return self._address_scope

    @property
    def original(self):
        return self._original_address_scope



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.nfp.lib import transport
import mock
from neutron.common import rpc as n_rpc
from neutron import context as ctx
from oslo_config import cfg
from oslo_serialization import jsonutils
import unittest

"""
Common class used to create configuration mapping
"""


class Map(dict):

    def __init__(self, *args, **kwargs):
        super(Map, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.iteritems():
                    self[k] = v

        if kwargs:
            for k, v in kwargs.iteritems():
                self[k] = v

    def __getattr__(self, attr):
        return self.get(attr)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(Map, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(Map, self).__delitem__(key)
        del self.__dict__[key]


class TestContext(object):

    def get_context(self):
        try:
            return ctx.Context('some_user', 'some_tenant')
        except Exception:
            return ctx.Context('some_user', 'some_tenant')

    def get_test_context(self):
        # creating a test context
        variables = {}
        variables['context'] = self.get_context()
        variables['body'] = {'info': {'context': {}},
                             'config': []}
        variables['method_type'] = 'CREATE'
        variables['device_config'] = True
        return variables


class CommonLibraryTest(unittest.TestCase):

    def setUp(self):
        n_rpc.init(cfg.CONF)
        self.imprt_rc = 'gbpservice.nfp.lib.rest_client_over_unix'

    def _cast(self, context, method, **kwargs):
        return

    def _call(self, context, method, **kwargs):
        return []

    def _get(self, path):

        class MockResponse(object):

            def __init__(self):
                self.content = {'success': '200'}
        return MockResponse()

    def _uget(self, path):
        return(200, "")

    def _post(self, path, body, method_type):
        return (200, "")

    def _upost(self, path, body, delete=False):
        return (200, "")

    def test_rpc_send_request_to_configurator(self):

        with mock.patch('oslo_messaging.rpc.client._CallContext.cast') as cast:
            cast.side_effect = self._cast

            test_context = TestContext().get_test_context()
            conf = Map(backend='rpc', RPC=Map(topic='topic'))

            transport.send_request_to_configurator(
                conf,
                test_context['context'],
                test_context['body'],
                test_context['method_type'],
                test_context['device_config'])

    def test_tcp_rest_send_request_to_configurator(self):

        with mock.patch.object(transport.RestApi, 'post') as mock_post:
            mock_post.side_effect = self._post

            test_context = TestContext().get_test_context()
            conf = Map(backend='tcp_rest', RPC=Map(topic='topic'),
                       REST=Map(rest_server_ip='0.0.0.0',
                                rest_server_port=5672))

            transport.send_request_to_configurator(
                conf,
                test_context['context'],
                test_context['body'],
                test_context['method_type'],
                test_context['device_config'])

    def test_unix_rest_send_request_to_configurator(self):

        with mock.patch(self.imprt_rc + '.post') as mock_post:
            mock_post.side_effect = self._upost

            test_context = TestContext().get_test_context()
            conf = Map(backend='unix_rest')

            transport.send_request_to_configurator(
                conf,
                test_context['context'],
                test_context['body'],
                test_context['method_type'],
                test_context['device_config'])

    def test_tcp_rest_get_response_from_configurator(self):

        with mock.patch.object(transport.RestApi, 'get') as (
            mock_get), mock.patch.object(jsonutils, 'loads') as (
            mock_loads):
            mock_get.side_effect = self._get
            mock_loads.return_value = True

            conf = Map(backend='tcp_rest', RPC=Map(topic='topic'),
                       REST=Map(rest_server_ip='0.0.0.0',
                                rest_server_port=5672))

            transport.get_response_from_configurator(conf)

    def test_unix_rest_get_response_from_configurator(self):

        with mock.patch(self.imprt_rc + '.get') as (
            mock_get), mock.patch.object(jsonutils, 'loads') as (
            mock_loads):
            mock_get.side_effect = self._uget
            mock_loads.return_value = True

            conf = Map(backend='unix_rest')

            transport.get_response_from_configurator(conf)

if __name__ == '__main__':
    unittest.main()



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Group Policy Implicit Policy Driver (gpm_ipd_1)

Revision ID: 23b6c4d703c7
Revises: 53de98f7a066
Create Date: 2014-07-24 15:33:34.751659

"""

# revision identifiers, used by Alembic.
revision = '23b6c4d703c7'
down_revision = '53de98f7a066'


from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'gpm_owned_l3_policies',
        sa.Column('l3_policy_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['l3_policy_id'], ['gp_l3_policies.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('l3_policy_id')
    )

    op.create_table(
        'gpm_owned_l2_policies',
        sa.Column('l2_policy_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['l2_policy_id'], ['gp_l2_policies.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('l2_policy_id')
    )


def downgrade():

    op.drop_table('gpm_owned_l2_policies')
    op.drop_table('gpm_owned_l3_policies')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.nfp.proxy_agent.modules import proxy_agent
import mock
from neutron import context as ctx
import unittest

rpc_manager = proxy_agent.RpcHandler


class TestContext(object):

    def get_context(self):
        try:
            return ctx.Context('some_user', 'some_tenant')
        except Exception:
            return ctx.Context('some_user', 'some_tenant')

"Common class for proxy agent test cases"


class ConfigAgentProxyTestCase(unittest.TestCase):

    def setUp(self):
        self.manager = rpc_manager('conf', 'sc')
        self.context = TestContext().get_context()
        self.import_lib = 'gbpservice.nfp.lib.transport'

    def _post(self, conf, context, body,
              method_type, device_config=False,
              network_function_event=False):
        return (200, '')

    def test_create_network_function_config(self):
        _data = "data"
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(import_send) as mock_send:
            mock_send.side_effect = self._post
            self.manager.create_network_function_config(self.context, _data)

    def test_delete_network_function_config(self):
        _data = "data"
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(import_send) as mock_send:
            mock_send.side_effect = self._post
            self.manager.delete_network_function_config(self.context, _data)

    def test_create_network_function_device_config(self):
        _data = "data"
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(import_send) as mock_send:
            mock_send.side_effect = self._post
            self.manager.create_network_function_device_config(
                self.context, _data)

    def test_delete_network_function_device_config(self):
        _data = "data"
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(import_send) as mock_send:
            mock_send.side_effect = self._post
            self.manager.delete_network_function_device_config(
                self.context, _data)

    def test_network_function_event(self):
        _data = "data"
        import_send = self.import_lib + '.send_request_to_configurator'
        with mock.patch(import_send) as mock_send:
            mock_send.side_effect = self._post
            self.manager.network_function_event(
                self.context, _data)

if __name__ == "__main__":
    unittest.main()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast

from neutron._i18n import _LE
from neutron.db import common_db_mixin
from neutron.db import model_base
from neutron.db import models_v2
from neutron import manager
from neutron.plugins.common import constants as pconst
from oslo_log import helpers as log
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy import orm
from sqlalchemy.orm import exc

from gbpservice.neutron.extensions import servicechain as schain
from gbpservice.neutron.services.servicechain.common import exceptions as s_exc


LOG = logging.getLogger(__name__)
MAX_IPV4_SUBNET_PREFIX_LENGTH = 31
MAX_IPV6_SUBNET_PREFIX_LENGTH = 127


class BaseSCResource(models_v2.HasId, models_v2.HasTenant):
    name = sa.Column(sa.String(255))
    description = sa.Column(sa.String(255))
    status = sa.Column(sa.String(length=16), nullable=True)
    status_details = sa.Column(sa.String(length=4096), nullable=True)


class BaseSharedSCResource(BaseSCResource):
    shared = sa.Column(sa.Boolean)


class SpecNodeAssociation(model_base.BASEV2):
    """Models one to many providing relation between Specs and Nodes."""
    __tablename__ = 'sc_spec_node_associations'
    servicechain_spec_id = sa.Column(
        sa.String(36), sa.ForeignKey('sc_specs.id'), primary_key=True)
    node_id = sa.Column(sa.String(36),
                        sa.ForeignKey('sc_nodes.id'),
                        primary_key=True)
    position = sa.Column(sa.Integer)


class InstanceSpecAssociation(model_base.BASEV2):
    """Models  one to many providing relation between Instance and Specs."""
    __tablename__ = 'sc_instance_spec_mappings'
    servicechain_instance_id = sa.Column(
        sa.String(36), sa.ForeignKey('sc_instances.id'), primary_key=True)
    servicechain_spec_id = sa.Column(sa.String(36),
                                     sa.ForeignKey('sc_specs.id'),
                                     primary_key=True)
    position = sa.Column(sa.Integer)


class ServiceChainNode(model_base.BASEV2, BaseSharedSCResource):
    """ServiceChain Node"""
    __tablename__ = 'sc_nodes'
    config = sa.Column(sa.TEXT)
    specs = orm.relationship(SpecNodeAssociation,
                             backref="nodes",
                             cascade='all, delete, delete-orphan')
    service_type = sa.Column(sa.String(50), nullable=True)
    service_profile_id = sa.Column(
        sa.String(36), sa.ForeignKey('service_profiles.id'),
        nullable=True)


class ServiceChainInstance(model_base.BASEV2, BaseSCResource):
    """Service chain instances"""
    __tablename__ = 'sc_instances'
    config_param_values = sa.Column(sa.String(4096))
    specs = orm.relationship(
        InstanceSpecAssociation,
        backref='instances',
        cascade='all,delete, delete-orphan',
        order_by='InstanceSpecAssociation.position',
        collection_class=ordering_list('position', count_from=1))
    provider_ptg_id = sa.Column(sa.String(36),
                             # FixMe(Magesh) Issue with cascade on Delete
                             # sa.ForeignKey('gp_policy_target_groups.id'),
                             nullable=True)
    consumer_ptg_id = sa.Column(sa.String(36),
                             # sa.ForeignKey('gp_policy_target_groups.id'),
                             nullable=True)
    management_ptg_id = sa.Column(sa.String(36),
                                  # sa.ForeignKey('gp_policy_target_groups.id'),
                                  nullable=True)
    classifier_id = sa.Column(sa.String(36),
                              # sa.ForeignKey('gp_policy_classifiers.id'),
                              nullable=True)


class ServiceChainSpec(model_base.BASEV2, BaseSharedSCResource):
    """ ServiceChain Spec
    """
    __tablename__ = 'sc_specs'
    nodes = orm.relationship(
        SpecNodeAssociation,
        backref='specs', cascade='all, delete, delete-orphan',
        order_by='SpecNodeAssociation.position',
        collection_class=ordering_list('position', count_from=1))
    config_param_names = sa.Column(sa.String(4096))
    instances = orm.relationship(InstanceSpecAssociation,
                                 backref="specs",
                                 cascade='all, delete, delete-orphan')


class ServiceProfile(model_base.BASEV2, BaseSharedSCResource):
    """ Service Profile
    """
    __tablename__ = 'service_profiles'
    vendor = sa.Column(sa.String(50))
    # Not using ENUM for less painful upgrades. Validation will happen at the
    # API level
    insertion_mode = sa.Column(sa.String(50))
    service_type = sa.Column(sa.String(50))
    service_flavor = sa.Column(sa.String(1024))
    nodes = orm.relationship(ServiceChainNode, backref="service_profile")


class ServiceChainDbPlugin(schain.ServiceChainPluginBase,
                           common_db_mixin.CommonDbMixin):
    """ServiceChain plugin interface implementation using SQLAlchemy models."""

    # TODO(osms69): native bulk support
    __native_bulk_support = False
    __native_pagination_support = True
    __native_sorting_support = True

    def __init__(self, *args, **kwargs):
        super(ServiceChainDbPlugin, self).__init__(*args, **kwargs)

    @property
    def _grouppolicy_plugin(self):
        # REVISIT(Magesh): Need initialization method after all
        # plugins are loaded to grab and store plugin.
        plugins = manager.NeutronManager.get_service_plugins()
        grouppolicy_plugin = plugins.get(pconst.GROUP_POLICY)
        if not grouppolicy_plugin:
            LOG.error(_LE("No Grouppolicy service plugin found."))
            raise s_exc.ServiceChainDeploymentError()
        return grouppolicy_plugin

    def _get_servicechain_node(self, context, node_id):
        try:
            return self._get_by_id(context, ServiceChainNode, node_id)
        except exc.NoResultFound:
            raise schain.ServiceChainNodeNotFound(sc_node_id=node_id)

    def _get_servicechain_spec(self, context, spec_id):
        try:
            return self._get_by_id(context, ServiceChainSpec, spec_id)
        except exc.NoResultFound:
            raise schain.ServiceChainSpecNotFound(sc_spec_id=spec_id)

    def _get_servicechain_instance(self, context, instance_id):
        try:
            return self._get_by_id(context, ServiceChainInstance, instance_id)
        except exc.NoResultFound:
            raise schain.ServiceChainInstanceNotFound(
                sc_instance_id=instance_id)

    def _get_service_profile(self, context, profile_id):
        try:
            return self._get_by_id(context, ServiceProfile, profile_id)
        except exc.NoResultFound:
            raise schain.ServiceProfileNotFound(
                profile_id=profile_id)

    def _populate_common_fields_in_dict(self, db_ref):
        res = {'id': db_ref['id'],
               'tenant_id': db_ref['tenant_id'],
               'name': db_ref['name'],
               'description': db_ref['description'],
               'status': db_ref['status'],
               'status_details': db_ref['status_details'],
               'shared': db_ref.get('shared', False)}
        return res

    def _make_sc_node_dict(self, sc_node, fields=None):
        res = self._populate_common_fields_in_dict(sc_node)
        res['service_profile_id'] = sc_node['service_profile_id']
        res['service_type'] = sc_node['service_type']
        res['config'] = sc_node['config']
        res['servicechain_specs'] = [sc_spec['servicechain_spec_id']
                                     for sc_spec in sc_node['specs']]
        return self._fields(res, fields)

    def _make_sc_spec_dict(self, spec, fields=None):
        res = self._populate_common_fields_in_dict(spec)
        res['config_param_names'] = spec.get('config_param_names')
        res['nodes'] = [sc_node['node_id'] for sc_node in spec['nodes']]
        res['instances'] = [x['servicechain_instance_id'] for x in
                            spec['instances']]
        return self._fields(res, fields)

    def _make_sc_instance_dict(self, instance, fields=None):
        res = {'id': instance['id'],
               'tenant_id': instance['tenant_id'],
               'name': instance['name'],
               'description': instance['description'],
               'config_param_values': instance['config_param_values'],
               'provider_ptg_id': instance['provider_ptg_id'],
               'consumer_ptg_id': instance['consumer_ptg_id'],
               'management_ptg_id': instance['management_ptg_id'],
               'classifier_id': instance['classifier_id'],
               'status': instance['status'],
               'status_details': instance['status_details']}
        res['servicechain_specs'] = [sc_spec['servicechain_spec_id']
                                    for sc_spec in instance['specs']]
        return self._fields(res, fields)

    def _make_service_profile_dict(self, profile, fields=None):
        res = self._populate_common_fields_in_dict(profile)
        res['service_type'] = profile['service_type']
        res['service_flavor'] = profile['service_flavor']
        res['vendor'] = profile['vendor']
        res['insertion_mode'] = profile['insertion_mode']
        res['nodes'] = [node['id'] for node in profile['nodes']]
        return self._fields(res, fields)

    @staticmethod
    def validate_service_type(service_type):
        if service_type not in schain.sc_supported_type:
            raise schain.ServiceTypeNotSupported(sc_service_type=service_type)

    @log.log_method_call
    def create_servicechain_node(self, context, servicechain_node):
        node = servicechain_node['servicechain_node']
        tenant_id = self._get_tenant_id_for_create(context, node)
        with context.session.begin(subtransactions=True):
            node_db = ServiceChainNode(
                id=uuidutils.generate_uuid(), tenant_id=tenant_id,
                name=node['name'], description=node['description'],
                service_profile_id=node.get('service_profile_id'),
                service_type=node.get('service_type'),
                config=node['config'], shared=node['shared'],
                status=node.get('status'),
                status_details=node.get('status_details'))
            context.session.add(node_db)
        return self._make_sc_node_dict(node_db)

    @log.log_method_call
    def update_servicechain_node(self, context, servicechain_node_id,
                                 servicechain_node, set_params=False):
        node = servicechain_node['servicechain_node']
        with context.session.begin(subtransactions=True):
            node_db = self._get_servicechain_node(context,
                                                  servicechain_node_id)
            node_db.update(node)
            # Update the config param names derived for the associated specs
            spec_node_associations = node_db.specs
            for node_spec in spec_node_associations:
                spec_id = node_spec.servicechain_spec_id
                spec_db = self._get_servicechain_spec(context, spec_id)
                self._process_nodes_for_spec(
                    context, spec_db, self._make_sc_spec_dict(spec_db),
                    set_params=set_params)
        return self._make_sc_node_dict(node_db)

    @log.log_method_call
    def delete_servicechain_node(self, context, servicechain_node_id):
        with context.session.begin(subtransactions=True):
            node_db = self._get_servicechain_node(context,
                                                  servicechain_node_id)
            if node_db.specs:
                raise schain.ServiceChainNodeInUse(
                                    node_id=servicechain_node_id)
            context.session.delete(node_db)

    @log.log_method_call
    def get_servicechain_node(self, context, servicechain_node_id,
                              fields=None):
        node = self._get_servicechain_node(context, servicechain_node_id)
        return self._make_sc_node_dict(node, fields)

    @log.log_method_call
    def get_servicechain_nodes(self, context, filters=None, fields=None,
                               sorts=None, limit=None, marker=None,
                               page_reverse=False):
        marker_obj = self._get_marker_obj(context, 'servicechain_node', limit,
                                          marker)
        return self._get_collection(context, ServiceChainNode,
                                    self._make_sc_node_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    @log.log_method_call
    def get_servicechain_nodes_count(self, context, filters=None):
        return self._get_collection_count(context, ServiceChainNode,
                                          filters=filters)

    def _process_nodes_for_spec(self, context, spec_db, spec,
                                set_params=True):
        if 'nodes' in spec:
            self._set_nodes_for_spec(context, spec_db, spec['nodes'],
                                     set_params=set_params)
            del spec['nodes']
        return spec

    def _set_nodes_for_spec(self, context, spec_db, nodes_id_list,
                            set_params=True):
        if not nodes_id_list:
            spec_db.nodes = []
            spec_db.config_param_names = '[]'
            return
        with context.session.begin(subtransactions=True):
            # We will first check if the new list of nodes is valid
            filters = {'id': [n_id for n_id in nodes_id_list]}
            nodes_in_db = self._get_collection_query(context, ServiceChainNode,
                                                     filters=filters)
            nodes_list = [n_db['id'] for n_db in nodes_in_db]
            for node_id in nodes_id_list:
                if node_id not in nodes_list:
                    # If we find an invalid node id in the list we
                    # do not perform the update
                    raise schain.ServiceChainNodeNotFound(sc_node_id=node_id)
            # New list of nodes is valid so we will first reset the
            #  existing list and then add each node in order.
            # Note that the list could be empty in which case we interpret
            # it as clearing existing nodes.
            spec_db.nodes = []
            if set_params:
                spec_db.config_param_names = '[]'
            for node_id in nodes_id_list:
                if set_params:
                    sc_node = self.get_servicechain_node(context, node_id)
                    node_dict = jsonutils.loads(sc_node['config'])
                    config_params = (node_dict.get('parameters') or
                                     node_dict.get('Parameters'))
                    if config_params:
                        if not spec_db.config_param_names:
                            spec_db.config_param_names = str(
                                config_params.keys())
                        else:
                            config_param_names = ast.literal_eval(
                                spec_db.config_param_names)
                            config_param_names.extend(config_params.keys())
                            spec_db.config_param_names = str(
                                config_param_names)

                assoc = SpecNodeAssociation(servicechain_spec_id=spec_db.id,
                                            node_id=node_id)
                spec_db.nodes.append(assoc)

    def _process_specs_for_instance(self, context, instance_db, instance):
        if 'servicechain_specs' in instance:
            self._set_specs_for_instance(context, instance_db,
                                         instance['servicechain_specs'])
            del instance['servicechain_specs']
        return instance

    def _set_specs_for_instance(self, context, instance_db, spec_id_list):
        if not spec_id_list:
            instance_db.spec_ids = []
            return
        with context.session.begin(subtransactions=True):
            filters = {'id': spec_id_list}
            specs_in_db = self._get_collection_query(context, ServiceChainSpec,
                                                     filters=filters)
            specs_list = set(spec_db['id'] for spec_db in specs_in_db)
            for spec_id in spec_id_list:
                if spec_id not in specs_list:
                    # Do not update if spec ID is invalid
                    raise schain.ServiceChainSpecNotFound(sc_spec_id=spec_id)
            # Reset the existing list and then add each spec in order. The list
            # could be empty in which case we clear the existing specs.
            instance_db.specs = []
            for spec_id in spec_id_list:
                assoc = InstanceSpecAssociation(
                                    servicechain_instance_id=instance_db.id,
                                    servicechain_spec_id=spec_id)
                instance_db.specs.append(assoc)

    def _get_instances_from_policy_target(self, context, policy_target):
        with context.session.begin(subtransactions=True):
            ptg_id = policy_target['policy_target_group_id']
            scis_p = self.get_servicechain_instances(
                context, {'provider_ptg_id': [ptg_id]})
            scis_c = self.get_servicechain_instances(
                context, {'consumer_ptg_id': [ptg_id]})
            # Don't return duplicates
            result = []
            seen = set()
            for sci in scis_p + scis_c:
                if sci['id'] not in seen:
                    seen.add(sci['id'])
                    result.append(sci)
            return result

    @log.log_method_call
    def create_servicechain_spec(self, context, servicechain_spec,
                                 set_params=True):
        spec = servicechain_spec['servicechain_spec']
        tenant_id = self._get_tenant_id_for_create(context, spec)
        with context.session.begin(subtransactions=True):
            spec_db = ServiceChainSpec(id=uuidutils.generate_uuid(),
                                       tenant_id=tenant_id,
                                       name=spec['name'],
                                       description=spec['description'],
                                       shared=spec['shared'],
                                       status=spec.get('status'),
                                       status_details=
                                       spec.get('status_details'))
            self._process_nodes_for_spec(context, spec_db, spec,
                                         set_params=set_params)
            context.session.add(spec_db)
        return self._make_sc_spec_dict(spec_db)

    @log.log_method_call
    def update_servicechain_spec(self, context, spec_id,
                                 servicechain_spec, set_params=True):
        spec = servicechain_spec['servicechain_spec']
        with context.session.begin(subtransactions=True):
            spec_db = self._get_servicechain_spec(context,
                                                  spec_id)
            spec = self._process_nodes_for_spec(context, spec_db, spec,
                                                set_params=set_params)
            spec_db.update(spec)
        return self._make_sc_spec_dict(spec_db)

    @log.log_method_call
    def delete_servicechain_spec(self, context, spec_id):
        policy_actions = self._grouppolicy_plugin.get_policy_actions(
                                context, filters={"action_value": [spec_id]})
        if policy_actions:
            raise schain.ServiceChainSpecInUse(spec_id=spec_id)
        with context.session.begin(subtransactions=True):
            spec_db = self._get_servicechain_spec(context,
                                                  spec_id)
            if spec_db.instances:
                raise schain.ServiceChainSpecInUse(spec_id=spec_id)
            context.session.delete(spec_db)

    @log.log_method_call
    def get_servicechain_spec(self, context, spec_id,
                              fields=None):
        spec = self._get_servicechain_spec(context, spec_id)
        return self._make_sc_spec_dict(spec, fields)

    @log.log_method_call
    def get_servicechain_specs(self, context, filters=None, fields=None,
                               sorts=None, limit=None, marker=None,
                               page_reverse=False):
        marker_obj = self._get_marker_obj(context, 'servicechain_spec', limit,
                                          marker)
        return self._get_collection(context, ServiceChainSpec,
                                    self._make_sc_spec_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    @log.log_method_call
    def get_servicechain_specs_count(self, context, filters=None):
        return self._get_collection_count(context, ServiceChainSpec,
                                          filters=filters)

    @log.log_method_call
    def create_servicechain_instance(self, context, servicechain_instance):
        instance = servicechain_instance['servicechain_instance']
        tenant_id = self._get_tenant_id_for_create(context, instance)
        with context.session.begin(subtransactions=True):
            if not instance.get('management_ptg_id'):
                management_groups = (
                    self._grouppolicy_plugin.get_policy_target_groups(
                        context, {'service_management': [True],
                                  'tenant_id': [instance.get('tenant_id')]}))
                if not management_groups:
                    # Fall back on shared service management
                    management_groups = (
                        self._grouppolicy_plugin.get_policy_target_groups(
                            context, {'service_management': [True]}))
                if management_groups:
                    instance['management_ptg_id'] = management_groups[0]['id']
            instance_db = ServiceChainInstance(
                id=uuidutils.generate_uuid(),
                tenant_id=tenant_id, name=instance['name'],
                description=instance['description'],
                config_param_values=instance['config_param_values'],
                provider_ptg_id=instance.get('provider_ptg_id'),
                consumer_ptg_id=instance.get('consumer_ptg_id'),
                management_ptg_id=instance.get('management_ptg_id'),
                classifier_id=instance.get('classifier_id'),
                status=instance.get('status'),
                status_details=instance.get('status_details'))
            self._process_specs_for_instance(context, instance_db, instance)
            context.session.add(instance_db)
        return self._make_sc_instance_dict(instance_db)

    @log.log_method_call
    def update_servicechain_instance(self, context, servicechain_instance_id,
                                     servicechain_instance):
        instance = servicechain_instance['servicechain_instance']
        with context.session.begin(subtransactions=True):
            instance_db = self._get_servicechain_instance(
                context, servicechain_instance_id)
            instance = self._process_specs_for_instance(context, instance_db,
                                                        instance)
            instance_db.update(instance)
        return self._make_sc_instance_dict(instance_db)

    @log.log_method_call
    def delete_servicechain_instance(self, context, servicechain_instance_id):
        with context.session.begin(subtransactions=True):
            instance_db = self._get_servicechain_instance(
                context, servicechain_instance_id)
            context.session.delete(instance_db)

    @log.log_method_call
    def get_servicechain_instance(self, context, sc_instance_id, fields=None):
        instance_db = self._get_servicechain_instance(context, sc_instance_id)
        return self._make_sc_instance_dict(instance_db, fields)

    @log.log_method_call
    def get_servicechain_instances(self, context, filters=None, fields=None,
                                   sorts=None, limit=None, marker=None,
                                   page_reverse=False):
        marker_obj = self._get_marker_obj(context, 'servicechain_instance',
                                          limit, marker)
        return self._get_collection(context, ServiceChainInstance,
                                    self._make_sc_instance_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker_obj,
                                    page_reverse=page_reverse)

    @log.log_method_call
    def get_servicechain_instances_count(self, context, filters=None):
        return self._get_collection_count(context, ServiceChainInstance,
                                          filters=filters)

    @log.log_method_call
    def get_service_profiles_count(self, context, filters=None):
        return self._get_collection_count(context, ServiceProfile,
                                          filters=filters)

    @log.log_method_call
    def create_service_profile(self, context, service_profile):
        profile = service_profile['service_profile']
        tenant_id = self._get_tenant_id_for_create(context, profile)
        with context.session.begin(subtransactions=True):
            profile_db = ServiceProfile(
                id=uuidutils.generate_uuid(), tenant_id=tenant_id,
                name=profile['name'], description=profile['description'],
                service_type=profile.get('service_type'),
                insertion_mode=profile.get('insertion_mode'),
                vendor=profile.get('vendor'),
                service_flavor=profile.get('service_flavor'),
                shared=profile.get('shared'),
                status=profile.get('status'),
                status_details=profile.get('status_details'))
            context.session.add(profile_db)
        return self._make_service_profile_dict(profile_db)

    @log.log_method_call
    def update_service_profile(self, context, service_profile_id,
                               service_profile):
        profile = service_profile['service_profile']
        with context.session.begin(subtransactions=True):
            profile_db = self._get_service_profile(context,
                                                   service_profile_id)
            profile_db.update(profile)
        return self._make_service_profile_dict(profile_db)

    @log.log_method_call
    def delete_service_profile(self, context, service_profile_id):
        with context.session.begin(subtransactions=True):
            profile_db = self._get_service_profile(context,
                                                   service_profile_id)
            if profile_db.nodes:
                raise schain.ServiceProfileInUse(
                    profile_id=service_profile_id)
            context.session.delete(profile_db)

    @log.log_method_call
    def get_service_profile(self, context, service_profile_id, fields=None):
        profile_db = self._get_service_profile(
            context, service_profile_id)
        return self._make_service_profile_dict(profile_db, fields)

    @log.log_method_call
    def get_service_profiles(self, context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        marker_obj = self._get_marker_obj(context, 'service_profile',
                                          limit, marker)
        return self._get_collection(context, ServiceProfile,
                                    self._make_service_profile_dict,
                                    filters=filters, fields=fields,
                                    sorts=sorts, limit=limit,
                                    marker_obj=marker_obj,
                                    page_reverse=page_reverse)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import copy

from octavia.common import constants
from octavia.common import data_models as o_data_models
from octavia.network import data_models as network_data_models

from gbpservice.contrib.nfp.configurator.drivers.base import base_driver
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.\
    v2.common import neutron_lbaas_data_models as n_data_models
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.\
    v2.haproxy import haproxy_driver_constants
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.\
    v2.haproxy.local_cert_manager import LocalCertManager
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.\
    v2.haproxy.rest_api_driver import HaproxyAmphoraLoadBalancerDriver
from gbpservice.contrib.nfp.configurator.lib import constants as common_const
from gbpservice.contrib.nfp.configurator.lib import data_parser
from gbpservice.contrib.nfp.configurator.lib import lb_constants
from gbpservice.contrib.nfp.configurator.lib import lbv2_constants
from gbpservice.nfp.common import exceptions
from gbpservice.nfp.core import log as nfp_logging


LOG = nfp_logging.getLogger(__name__)


# Copy from loadbalancer/v1/haproxy/haproxy_lb_driver.py
""" Loadbalancer generic configuration driver for handling device
configuration requests.
"""


class LbGenericConfigDriver(object):
    """
    Driver class for implementing loadbalancer configuration
    requests from Orchestrator.
    """

    def __init__(self):
        pass

    def configure_interfaces(self, context, resource_data):
        """ Configure interfaces for the service VM.
        Calls static IP configuration function and implements
        persistent rule addition in the service VM.
        Issues REST call to service VM for configuration of interfaces.
        :param context: neutron context
        :param resource_data: a dictionary of loadbalancer objects
        send by neutron plugin
        Returns: SUCCESS/Failure message with reason.
        """
        resource_data = self.parse.parse_data(
            common_const.INTERFACES, resource_data)
        mgmt_ip = resource_data['mgmt_ip']

        try:
            result_log_forward = self._configure_log_forwarding(
                lb_constants.REQUEST_URL, mgmt_ip,
                self.port)
        except Exception as err:
            msg = ("Failed to configure log forwarding for service at %s. "
                   "Error: %s" % (mgmt_ip, err))
            LOG.error(msg)
            return msg
        else:
            if result_log_forward == common_const.UNHANDLED:
                pass
            elif result_log_forward != lb_constants.STATUS_SUCCESS:
                msg = ("Failed to configure log forwarding for service at %s. "
                       % mgmt_ip)
                LOG.error(msg)
                return result_log_forward
            else:
                msg = ("Configured log forwarding for service at %s. "
                       "Result: %s" % (mgmt_ip, result_log_forward))
                LOG.info(msg)

        return lb_constants.STATUS_SUCCESS


# As we use the rest client and amphora image from Octavia,
# we need to have a helper class to simulate Octavia DB operation
# in order to get Octavia data models from Neutron-lbaas data models
class OctaviaDataModelBuilder(object):

    def __init__(self, driver=None):
        self.driver = driver

    # All Octavia data models have these attributes
    def _get_common_args(self, obj):
        return {
            'id': obj.id,
            'project_id': obj.tenant_id,
            'name': obj.name,
            'description': obj.description,
            'enabled': obj.admin_state_up,
            'operating_status': obj.operating_status,
        }

    # Update Octavia model from dict
    def _update(self, octavia_data_model, update_dict):
        for key, value in update_dict.items():
            setattr(octavia_data_model, key, value)
        return octavia_data_model

    # Translate loadbalancer neutron model dict to octavia model
    def get_loadbalancer_octavia_model(self, loadbalancer_dict):
        loadbalancer = n_data_models.LoadBalancer.from_dict(
            copy.deepcopy(loadbalancer_dict))
        ret = o_data_models.LoadBalancer()
        args = self._get_common_args(loadbalancer)
        vip = o_data_models.Vip(
            load_balancer_id=loadbalancer.id,
            ip_address=loadbalancer.vip_address,
            subnet_id=loadbalancer.vip_subnet_id,
            port_id=loadbalancer.vip_port.id,
            load_balancer=ret
        )
        amphorae = self.driver.get_amphora(loadbalancer.id)
        if not amphorae:
            raise exceptions.IncompleteData(
                "Amphora information is missing")
        # REVISIT(jiahao): cluster_group, topology, affinity_group_id are not
        # included yet
        args.update({
            'vip': vip,
            'amphorae': amphorae,
            'provisioning_status': loadbalancer.provisioning_status,
        })
        if loadbalancer_dict.get('listeners'):
            listeners = []
            pools = []
            for listener_dict in loadbalancer_dict.get('listeners'):
                listener = self.get_listener_octavia_model(listener_dict)
                listener.load_balancer = ret
                listeners.append(listener)
                pools.extend(listener.pools)
                for pool in listener.pools:
                    if pool.id not in [pool.id for pool in pools]:
                        pools.append(pool)
            args.update({
                'listeners': listeners,
                'pools': pools,
            })

        ret = self._update(ret, args)
        return ret

    # Translate listener neutron model dict to octavia model
    def get_listener_octavia_model(self, listener_dict):
        # Must use a copy because from_dict will modify the original dict
        listener = n_data_models.Listener.from_dict(
            copy.deepcopy(listener_dict))
        ret = o_data_models.Listener()
        args = self._get_common_args(listener)
        sni_containers = []
        if listener_dict.get('sni_containers'):
            for sni_dict in listener_dict.get('sni_containers'):
                sni = o_data_models.SNI()
                if sni_dict.get('listener'):
                    sni.listener = self.get_listener_octavia_model(
                        sni_dict.get('listener'))
                sni.listener_id = sni_dict.get('listener_id')
                sni.position = sni_dict.get('position')
                sni.tls_container_id = sni_dict.get('tls_container_id')
                sni_containers.append(sni)
        if listener_dict.get('loadbalancer'):
            loadbalancer = self.get_loadbalancer_octavia_model(
                listener_dict.get('loadbalancer'))
            if listener.id not in [_listener.id for _listener
                                   in loadbalancer.listeners]:
                loadbalancer.listeners.append(ret)
            args.update({
                'load_balancer': loadbalancer,
            })
        if listener_dict.get('default_pool'):
            pool = self.get_pool_octavia_model(
                listener_dict.get('default_pool'))
            if listener.id not in [_listener.id for _listener
                                   in pool.listeners]:
                pool.listeners.append(ret)
            # REVISIT(jiahao): In Mitaka, we need to handle multiple pools
            pools = [pool]
            args.update({
                'default_pool': pool,
                'pools': pools,
            })
        args.update({
            'load_balancer_id': listener.loadbalancer_id,
            'protocol': listener.protocol,
            'protocol_port': listener.protocol_port,
            'connection_limit': listener.connection_limit,
            'default_pool_id': listener.default_pool_id,
            'tls_certificate_id': listener.default_tls_container_id,
            'sni_containers': sni_containers,
            'provisioning_status': listener.provisioning_status,
        })
        ret = self._update(ret, args)
        return ret

    # Translate pool neutron model dict to octavia model
    def get_pool_octavia_model(self, pool_dict):
        pool = n_data_models.Pool.from_dict(
            copy.deepcopy(pool_dict)
        )
        ret = o_data_models.Pool()
        args = self._get_common_args(pool)
        # REVISIT(jiahao): In Mitaka, instead of pool.listener,
        # there are pool.listeners. We need to handle that
        if pool_dict.get('listener'):
            listener = self.get_listener_octavia_model(
                pool_dict.get('listener'))
            if pool.id not in [_pool.id for _pool in listener.pools]:
                listener.pools.append(ret)
            if (not listener.default_pool) \
                    or (listener.default_pool_id == pool.id):
                listener.default_pool = ret
            listeners = [listener]
            args.update({
                'listeners': listeners,
            })
            if listener.load_balancer:
                if pool.id not in [_pool.id for _pool
                                   in listener.load_balancer.pools]:
                    listener.load_balancer.pools.append(ret)
                args.update({
                    'load_balancer': listener.load_balancer,
                    'load_balancer_id': listener.load_balancer_id,
                })
        if pool_dict.get('members'):
            members = []
            for member_dict in pool_dict.get('members'):
                member = self.get_member_octavia_model(member_dict)
                if not member.pool:
                    member.pool = ret
                members.append(member)
            args.update({
                'members': members
            })
        if pool_dict.get('healthmonitor'):
            healthmonitor = self.get_healthmonitor_octavia_model(
                pool_dict.get('healthmonitor'))
            if not healthmonitor.pool:
                healthmonitor.pool = ret
            args.update({
                'health_monitor': healthmonitor
            })

        # REVISIT(jiahao): L7Policy are not added
        args.update({
            'protocol': pool.protocol,
            'lb_algorithm': pool.lb_algorithm,
            'session_persistence': pool.session_persistence,
        })
        ret = self._update(ret, args)
        return ret

    # Translate member neutron model dict to octavia model
    def get_member_octavia_model(self, member_dict):
        member = n_data_models.Member.from_dict(
            copy.deepcopy(member_dict)
        )
        ret = o_data_models.Member()
        args = {
            'id': member.id,
            'project_id': member.tenant_id,
            'pool_id': member.pool_id,
            'ip_address': member.address,
            'protocol_port': member.protocol_port,
            'weight': member.weight,
            'enabled': member.admin_state_up,
            'subnet_id': member.subnet_id,
            'operating_status': member.operating_status,
        }
        if member_dict.get('pool'):
            pool = self.get_pool_octavia_model(member_dict.get('pool'))
            args.update({
                'pool': pool
            })
        ret = self._update(ret, args)
        return ret

    # Translate HealthMonitor neutron model dict to octavia model
    def get_healthmonitor_octavia_model(self, hm_dict):
        hm = n_data_models.HealthMonitor.from_dict(
            copy.deepcopy(hm_dict)
        )
        ret = o_data_models.HealthMonitor()
        args = {
            'id': hm.id,
            'project_id': hm.tenant_id,
            'type': hm.type,
            'delay': hm.delay,
            'timeout': hm.timeout,
            'rise_threshold': hm.max_retries,
            'fall_threshold': hm.max_retries,
            'http_method': hm.http_method,
            'url_path': hm.url_path,
            'expected_codes': hm.expected_codes,
            'enabled': hm.admin_state_up
        }
        if hm_dict.get('pool'):
            pool = self.get_pool_octavia_model(hm_dict.get('pool'))
            args.update({
                'pool': pool,
                'pool_id': pool.id
            })
        ret = self._update(ret, args)
        return ret


@base_driver.set_class_attr(
    SERVICE_TYPE=lbv2_constants.SERVICE_TYPE,
    SERVICE_VENDOR=haproxy_driver_constants.SERVICE_VENDOR)
class HaproxyLoadBalancerDriver(LbGenericConfigDriver,
                                base_driver.BaseDriver):

    # amphorae = {"loadbalancer_id": [o_data_models.Amphora(
    #                                 lb_network_ip, id, status)]}
    amphorae = {}

    def __init__(self, plugin_rpc=None, conf=None):
        # Each of the major LBaaS objects in the neutron database
        # need a corresponding manager/handler class.
        #
        # Put common things that are shared across the entire driver, like
        # config or a rest client handle, here.
        #
        # This function is executed when neutron-server starts.
        super(HaproxyLoadBalancerDriver, self).__init__()
        self.conf = conf
        self.port = haproxy_driver_constants.CONFIGURATION_SERVER_PORT
        self.parse = data_parser.DataParser()
        self.amphora_driver = HaproxyAmphoraLoadBalancerDriver()
        self.cert_manager = LocalCertManager()

        self.load_balancer = HaproxyLoadBalancerManager(self)
        self.listener = HaproxyListenerManager(self)
        self.pool = HaproxyPoolManager(self)
        self.member = HaproxyMemberManager(self)
        self.health_monitor = HaproxyHealthMonitorManager(self)
        self.o_models_builder = OctaviaDataModelBuilder(self)

    @classmethod
    def get_name(cls):
        return haproxy_driver_constants.DRIVER_NAME

    # Get Amphora object given the loadbalancer_id
    def get_amphora(self, loadbalancer_id):
        return self.amphorae.get(loadbalancer_id)

    def add_amphora(self, context, loadbalancer_id, description,
                    status=constants.ACTIVE):
        sc_metadata = ast.literal_eval(description)
        rdata = self.parse.parse_data(common_const.LOADBALANCERV2, context)
        if not (rdata['mgmt_ip'] and sc_metadata.get('network_function_id')):
            raise exceptions.IncompleteData(
                "Amphora information is missing")
        if not self.get_amphora(loadbalancer_id):
            # REVISIT(jiahao): use network_function_id as amphora id
            amp = o_data_models.Amphora(
                lb_network_ip=rdata['mgmt_ip'],
                id=sc_metadata['network_function_id'],
                status=status)
            self.amphorae[loadbalancer_id] = [amp]


class HaproxyCommonManager(object):

    def __init__(self, driver):
        self.driver = driver
        self.parse = data_parser.DataParser()

    def _deploy(self, context, obj):
        pass

    def create(self, context, obj):
        msg = ("LB %s, created %s" % (self.__class__.__name__, obj['id']))
        LOG.info(msg)

    def update(self, context, old_obj, obj):
        msg = ("LB %s, updated %s" % (self.__class__.__name__, obj['id']))
        LOG.info(msg)

    def delete(self, context, obj):
        msg = ("LB %s, deleted %s" % (self.__class__.__name__, obj['id']))
        LOG.info(msg)

    def store_certs(self, listener_obj, listener_dict):
        cert_mngr = self.driver.cert_manager
        cert_ids = []
        if listener_obj.tls_certificate_id:
            cert = listener_dict["default_tls_container"]
            tls_certificate_id = cert_mngr.store_cert(
                project_id=listener_dict["tenant_id"],
                certificate=cert["certificate"],
                private_key=cert["private_key"],
                intermediates=cert["intermediates"]
            )
            listener_obj.tls_certificate_id = tls_certificate_id
            cert_ids.append(tls_certificate_id)

        if listener_obj.sni_containers:
            for sni_cont in listener_obj.sni_containers:
                for cont in listener_dict["sni_containers"]:
                    if sni_cont.tls_container_id == cont["tls_container_id"]:
                        cert = cont["tls_container"]
                        tls_certificate_id = cert_mngr.store_cert(
                            project_id=listener_dict["tenant_id"],
                            certificate=cert["certificate"],
                            private_key=cert["private_key"],
                            intermediates=cert["intermediates"]
                        )
                        sni_cont.tls_container_id = tls_certificate_id
                        cert_ids.append(tls_certificate_id)
                        break

        return cert_ids

    def clean_certs(self, project_id, cert_ids):
        cert_mngr = self.driver.cert_manager
        for cert_id in cert_ids:
            cert_mngr.delete_cert(project_id, cert_id)


class HaproxyLoadBalancerManager(HaproxyCommonManager):

    def _get_amphorae_network_config(self,
                                     context,
                                     loadbalancer_dict,
                                     loadbalancer_o_obj):
        loadbalancer_n_obj = n_data_models.LoadBalancer.from_dict(
            copy.deepcopy(loadbalancer_dict))

        amphorae_network_config = {}

        for amp in loadbalancer_o_obj.amphorae:
            if amp.status != constants.DELETED:
                # Get vip_subnet
                vip_subnet = None
                for subnet_dict in context['service_info']['subnets']:
                    if subnet_dict['id'] == loadbalancer_n_obj.vip_subnet_id:
                        vip_subnet = n_data_models.Subnet.from_dict(
                            copy.deepcopy(subnet_dict))
                        break
                if vip_subnet is None:
                    raise exceptions.IncompleteData(
                        "VIP subnet information is not found")

                sc_metadata = self.parse.parse_data(
                    common_const.LOADBALANCERV2, context)
                vrrp_port = n_data_models.Port(
                    mac_address=sc_metadata['provider_mac'])
                if vrrp_port is None:
                    raise exceptions.IncompleteData(
                        "VRRP port information is not found")

                amphorae_network_config[amp.id] = \
                    network_data_models.AmphoraNetworkConfig(
                        amphora=amp,
                        vip_subnet=vip_subnet,
                        vrrp_port=vrrp_port)

        return amphorae_network_config

    def create(self, context, loadbalancer):
        self.driver.add_amphora(context, loadbalancer['id'],
                                loadbalancer['description'])
        loadbalancer_o_obj = self.driver.o_models_builder.\
            get_loadbalancer_octavia_model(loadbalancer)
        amphorae_network_config = self._get_amphorae_network_config(
            context, loadbalancer, loadbalancer_o_obj)
        for amp in loadbalancer_o_obj.amphorae:
            self.driver.amphora_driver.post_vip_plug(
                amp, loadbalancer_o_obj, amphorae_network_config)

        msg = ("LB %s, created %s"
               % (self.__class__.__name__, loadbalancer['id']))
        LOG.info(msg)
        msg = ("Notified amphora of vip plug. "
               "Loadbalancer id: %s, vip: %s"
               % (loadbalancer['id'], loadbalancer_o_obj.vip.ip_address))
        LOG.info(msg)

    def update(self, context, old_loadbalancer, loadbalancer):
        self.driver.add_amphora(context, loadbalancer['id'],
                                loadbalancer['description'])
        loadbalancer_o_obj = self.driver.o_models_builder.\
            get_loadbalancer_octavia_model(loadbalancer)
        for listener in loadbalancer_o_obj.listeners:
            cert_ids = []
            for listener_dict in loadbalancer['listeners']:
                if listener.id == listener_dict['id']:
                    cert_ids = self.store_certs(listener, listener_dict)
                    break
            self.driver.amphora_driver.update(listener, loadbalancer_o_obj.vip)
            self.clean_certs(loadbalancer['tenant_id'], cert_ids)

        msg = ("LB %s, updated %s"
               % (self.__class__.__name__, loadbalancer['id']))
        LOG.info(msg)

    def delete(self, context, loadbalancer):
        msg = ("LB %s, deleted %s"
               % (self.__class__.__name__, loadbalancer['id']))
        LOG.info(msg)
        # delete loadbalancer doesn't need any operation on service vm

    @property
    def allocates_vip(self):
        msg = ('allocates_vip queried')
        LOG.info(msg)
        return False

    def create_and_allocate_vip(self, context, loadbalancer):
        msg = ("LB %s, create_and_allocate_vip %s"
               % (self.__class__.__name__, loadbalancer['id']))
        LOG.info(msg)
        self.create(context, loadbalancer)

    def refresh(self, context, loadbalancer):
        # This is intended to trigger the backend to check and repair
        # the state of this load balancer and all of its dependent objects
        msg = ("LB pool refresh %s" % (loadbalancer['id']))
        LOG.info(msg)

    def stats(self, context, loadbalancer):
        msg = ("LB stats %s" % (loadbalancer['id']))
        LOG.info(msg)
        return {
            "bytes_in": 0,
            "bytes_out": 0,
            "active_connections": 0,
            "total_connections": 0
        }


class HaproxyListenerManager(HaproxyCommonManager):

    def _deploy(self, context, listener):
        self.driver.add_amphora(context, listener['loadbalancer_id'],
                                listener['description'])
        listener_o_obj = self.driver.o_models_builder.\
            get_listener_octavia_model(listener)
        cert_ids = self.store_certs(listener_o_obj, listener)
        self.driver.amphora_driver.update(listener_o_obj,
                                          listener_o_obj.load_balancer.vip)
        self.clean_certs(listener['tenant_id'], cert_ids)

    def create(self, context, listener):
        self._deploy(context, listener)
        msg = ("LB %s, created %s" % (self.__class__.__name__, listener['id']))
        LOG.info(msg)

    def update(self, context, old_listener, listener):
        self._deploy(context, listener)
        msg = ("LB %s, updated %s" % (self.__class__.__name__, listener['id']))
        LOG.info(msg)

    def delete(self, context, listener):
        self.driver.add_amphora(context, listener['loadbalancer_id'],
                                listener['description'])
        listener_o_obj = self.driver.o_models_builder.\
            get_listener_octavia_model(listener)
        self.driver.amphora_driver.delete(listener_o_obj,
                                          listener_o_obj.load_balancer.vip)
        msg = ("LB %s, deleted %s" % (self.__class__.__name__, listener['id']))
        LOG.info(msg)


class HaproxyPoolManager(HaproxyCommonManager):

    def _remove_pool(self, pool):
        pool_id = pool['id']
        # REVISIT(jiahao): In Mitaka, we need to handle multiple pools
        default_pool = pool['listener']['default_pool']
        if default_pool['id'] == pool_id:
            pool['listener']['default_pool'] = None

    def _deploy(self, context, pool):
        self.driver.add_amphora(context, pool['loadbalancer_id'],
                                pool['description'])
        pool_o_obj = self.driver.o_models_builder.\
            get_pool_octavia_model(pool)
        # For Mitaka, that would be multiple listeners within pool
        listener_o_obj = pool_o_obj.listeners[0]
        load_balancer_o_obj = pool_o_obj.load_balancer
        cert_ids = self.store_certs(listener_o_obj,
                                    pool['listeners'][0])
        self.driver.amphora_driver.update(listener_o_obj,
                                          load_balancer_o_obj.vip)
        self.clean_certs(pool['tenant_id'], cert_ids)

    def create(self, context, pool):
        self._deploy(context, pool)
        msg = ("LB %s, created %s" % (self.__class__.__name__, pool['id']))
        LOG.info(msg)

    def update(self, context, old_pool, pool):
        self._deploy(context, pool)
        msg = ("LB %s, updated %s" % (self.__class__.__name__, pool['id']))
        LOG.info(msg)

    def delete(self, context, pool):
        self._remove_pool(pool)
        self._deploy(context, pool)
        msg = ("LB %s, deleted %s" % (self.__class__.__name__, pool['id']))
        LOG.info(msg)


class HaproxyMemberManager(HaproxyCommonManager):

    def _deploy(self, context, member):
        self.driver.add_amphora(context, member['pool']['loadbalancer_id'],
                                member['description'])
        member_o_obj = self.driver.o_models_builder.\
            get_member_octavia_model(member)
        listener_o_obj = member_o_obj.pool.listeners[0]
        load_balancer_o_obj = member_o_obj.pool.load_balancer
        cert_ids = self.store_certs(listener_o_obj,
                                    member['pool']['listeners'][0])
        self.driver.amphora_driver.update(listener_o_obj,
                                          load_balancer_o_obj.vip)
        self.clean_certs(member['tenant_id'], cert_ids)

    def _remove_member(self, member):
        member_id = member['id']
        # REVISIT(jiahao): In Mitaka, we need to handle multiple pools
        default_pool = member['pool']['listener']['default_pool']
        for index, item in enumerate(default_pool['members']):
            if item['id'] == member_id:
                default_pool['members'].pop(index)
                break

    def create(self, context, member):
        self._deploy(context, member)
        msg = ("LB %s, created %s" % (self.__class__.__name__, member['id']))
        LOG.info(msg)

    def update(self, context, old_member, member):
        self._deploy(context, member)
        msg = ("LB %s, updated %s" % (self.__class__.__name__, member['id']))
        LOG.info(msg)

    def delete(self, context, member):
        self._remove_member(member)
        self._deploy(context, member)
        msg = ("LB %s, deleted %s" % (self.__class__.__name__, member['id']))
        LOG.info(msg)


class HaproxyHealthMonitorManager(HaproxyCommonManager):

    def _deploy(self, context, hm):
        self.driver.add_amphora(context, hm['pool']['loadbalancer_id'],
                                hm['description'])
        hm_o_obj = self.driver.o_models_builder.\
            get_healthmonitor_octavia_model(hm)
        listener_o_obj = hm_o_obj.pool.listeners[0]
        load_balancer_o_obj = hm_o_obj.pool.load_balancer
        cert_ids = self.store_certs(listener_o_obj,
                                    hm['pool']['listeners'][0])
        self.driver.amphora_driver.update(listener_o_obj,
                                          load_balancer_o_obj.vip)
        self.clean_certs(hm['tenant_id'], cert_ids)

    def _remove_healthmonitor(self, hm):
        hm_id = hm['id']
        default_pool = hm['pool']['listener']['default_pool']
        if default_pool['healthmonitor']['id'] == hm_id:
            default_pool['healthmonitor'] = None

    def create(self, context, hm):
        self._deploy(context, hm)
        msg = ("LB %s, created %s" % (self.__class__.__name__, hm['id']))
        LOG.info(msg)

    def update(self, context, old_hm, hm):
        self._deploy(context, hm)
        msg = ("LB %s, updated %s" % (self.__class__.__name__, hm['id']))
        LOG.info(msg)

    def delete(self, context, hm):
        self._remove_healthmonitor(hm)
        self._deploy(context, hm)
        msg = ("LB %s, deleted %s" % (self.__class__.__name__, hm['id']))
        LOG.info(msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from gbpservice.contrib.nfp.configurator.agents import loadbalancer_v1 as lb
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.v1.\
    haproxy import (haproxy_lb_driver as lb_driver)
from gbpservice.contrib.nfp.configurator.drivers.loadbalancer.v1.\
    haproxy import (haproxy_rest_client as _rest_client)
from gbpservice.contrib.tests.unit.nfp.configurator.test_data import (
    lb_test_data as test_data)
from neutron.tests import base
from oslo_serialization import jsonutils


class HaproxyOnVmDriverTestCase(base.BaseTestCase):
    """ Implements test cases for haproxy loadbalancer driver. """

    def __init__(self, *args, **kwargs):
        super(HaproxyOnVmDriverTestCase, self).__init__(*args, **kwargs)
        self.fo = test_data.FakeObjects()
        self.data = test_data.AssertionData()
        self.driver = lb_driver.HaproxyOnVmDriver()
        self.resp = mock.Mock()
        self.fake_resp_dict = {'status': True,
                               'config_success': True,
                               'delete_success': True}
        self.fo.vip = self.fo._get_vip_object()[0]
        self.fo.old_vip = self.fo._get_vip_object()[0]
        self.fo.pool = self.fo._get_pool_object()[0]
        self.fo.old_pool = self.fo._get_pool_object()[0]
        self.fo.hm = self.fo._get_hm_object()
        self.fo.old_hm = self.fo._get_hm_object()
        self.fo.member = self.fo._get_member_object()
        self.fo.old_member = self.fo._get_member_object()
        self.vip = self.fo.vip
        self.resp.status_code = 200
        self.get_resource = {
            'server': {
                'resource': [],
                'srvr:4910851f-4af7-4592-ad04-08b508c6fa21': []},
            'timeout': {}}

    @mock.patch(__name__ + '.test_data.FakeObjects.rpcmgr')
    @mock.patch(__name__ + '.test_data.FakeObjects.drivers')
    @mock.patch(__name__ + '.test_data.FakeObjects.sc')
    def _get_lb_handler_objects(self, sc, drivers, rpcmgr):
        """ Retrieves EventHandler object of loadbalancer agent.

        :param sc: mocked service controller object of process model framework
        :param drivers: mocked drivers object of loadbalancer object
        :param rpcmgr: mocked RPC manager object loadbalancer object

        Returns: objects of LBaaSEventHandler of loadbalancer agent

        """

        agent = lb.LBaaSEventHandler(sc, drivers, rpcmgr)
        return agent

    def _test_lbaasdriver(self, method_name):
        """ Tests all create/update/delete operation of loadbalancer driver.

        Returns: none

        """
        agent = self._get_lb_handler_objects()
        driver = lb_driver.HaproxyOnVmDriver(agent.plugin_rpc)
        rest_client = _rest_client.HttpRequests(
            self.data.url, self.data.port)
        logical_device_return_value = {
            'vip': self.fo.vip,
            'old_vip': self.fo.old_vip,
            'pool': self.fo.pool,
            'healthmonitors': self.fo.hm,
            'members': self.fo.member}
        with mock.patch.object(
                agent.plugin_rpc, 'get_logical_device',
                return_value=logical_device_return_value), (
            mock.patch.object(
                driver, '_get_rest_client', return_value=rest_client)), (
            mock.patch.object(
                rest_client.pool, 'request',
                return_value=self.resp)) as mock_request, (
            mock.patch.object(
                rest_client, 'get_resource',
                return_value=self.get_resource)) as mock_get_resource:

            mock_request.status_code = 200
            if method_name == 'DELETE_VIP':
                driver.delete_vip(self.fo.vip, self.fo.lb_api_context())
                mock_request.assert_called_with(
                    'DELETE',
                    data=None,
                    headers=self.data.header,
                    timeout=self.data.timeout,
                    url=self.data.delete_vip_url)
            elif method_name == 'CREATE_VIP':
                driver.create_vip(self.fo.vip, self.fo.lb_api_context())
                data = jsonutils.dumps(self.data.create_vip_data)
                mock_request.assert_called_with(
                    'POST',
                    data=data,
                    headers=self.data.header,
                    timeout=30,
                    url=self.data.create_vip_url)
                mock_get_resource.assert_called_with(
                    self.data.create_vip_resources)
            elif method_name == 'UPDATE_VIP':
                driver.update_vip(
                    self.fo.old_vip,
                    self.fo.vip,
                    self.fo.lb_api_context())
                data = jsonutils.dumps(self.data.update_vip_data)
                mock_request.assert_called_with(
                    'PUT',
                    data=data,
                    headers=self.data.header,
                    timeout=self.data.timeout,
                    url=self.data.update_vip_url)
            elif method_name == 'CREATE_POOL':
                driver.create_pool(self.fo.pool, self.fo.lb_api_context())
            elif method_name == 'DELETE_POOL':
                driver.delete_pool(self.fo.pool, self.fo.lb_api_context())
            elif method_name == 'UPDATE_POOL':
                driver.update_pool(
                    self.fo.old_pool,
                    self.fo.pool,
                    self.fo.lb_api_context())
                data = jsonutils.dumps(self.data.update_pool_data)
                mock_request.assert_called_with(
                    'PUT',
                    data=data,
                    headers=self.data.header,
                    timeout=self.data.timeout,
                    url=self.data.update_pool_url)
            elif method_name == 'CREATE_MEMBER':
                driver.create_member(self.fo.member[0],
                                     self.fo.lb_api_context())
                data = jsonutils.dumps(self.data.create_member_data)
                mock_request.assert_called_with(
                    'PUT',
                    data=data,
                    headers=self.data.header,
                    timeout=self.data.timeout,
                    url=self.data.create_member_url)
            elif method_name == 'DELETE_MEMBER':
                driver.delete_member(self.fo.member[0],
                                     self.fo.lb_api_context())
                data = jsonutils.dumps(self.data.delete_member_data)
                mock_request.assert_called_with(
                    'PUT',
                    data=data,
                    headers=self.data.header,
                    timeout=self.data.timeout,
                    url=self.data.delete_member_url)
            elif method_name == 'UPDATE_MEMBER':
                driver.update_member(
                    self.fo.old_member[0],
                    self.fo.member[0],
                    self.fo.lb_api_context())
                data = jsonutils.dumps(self.data.update_member_data)
                mock_request.assert_called_with(
                    'PUT',
                    data=data,
                    headers=self.data.header,
                    timeout=self.data.timeout,
                    url=self.data.update_member_url)
            elif method_name == 'CREATE_POOL_HEALTH_MONITOR':
                driver.create_pool_health_monitor(
                    self.fo.hm[0], self.fo._get_pool_object()[0]['id'],
                    self.fo.lb_api_context())
                data = jsonutils.dumps(self.data.create_hm_data)
                mock_request.assert_called_with(
                    'PUT',
                    data=data,
                    headers=self.data.header,
                    timeout=self.data.timeout,
                    url=self.data.create_hm_url)
            elif method_name == 'DELETE_POOL_HEALTH_MONITOR':
                driver.delete_pool_health_monitor(
                    self.fo.hm[0], self.fo._get_pool_object()[0]['id'],
                    self.fo.lb_api_context())
                data = jsonutils.dumps(self.data.delete_hm_data)
                mock_request.assert_called_with(
                    'PUT',
                    data=data,
                    headers=self.data.header,
                    timeout=self.data.timeout,
                    url=self.data.delete_hm_url)
            elif method_name == 'UPDATE_POOL_HEALTH_MONITOR':
                driver.update_pool_health_monitor(
                    self.fo.old_hm[0],
                    self.fo.hm[0], self.fo._get_pool_object()[0]['id'],
                    self.fo.lb_api_context())
                data = jsonutils.dumps(self.data.update_hm_data)
                mock_request.assert_called_with(
                    'PUT',
                    data=data,
                    headers=self.data.header,
                    timeout=self.data.timeout,
                    url=self.data.update_hm_url)

    def test_vip_create_lbaasdriver(self):
        """Implements test case for create vip method of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('CREATE_VIP')

    def test_vip_delete_lbaasdriver(self):
        """Implements test case for delete vip method of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('DELETE_VIP')

    def test_vip_update_lbaasdriver(self):
        """Implements test case for update vip method of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('UPDATE_VIP')

    def test_pool_create_lbaasdriver(self):
        """Implements test case for create pool method of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('CREATE_POOL')

    def test_pool_delete_lbaasdriver(self):
        """Implements test case for delete vip method of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('DELETE_POOL')

    def test_pool_update_lbaasdriver(self):
        """Implements test case for update vip method of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('UPDATE_POOL')

    def test_member_create_lbaasdriver(self):
        """Implements test case for create member method of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('CREATE_MEMBER')

    def test_member_delete_lbaasdriver(self):
        """Implements test case for delete member method of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('DELETE_MEMBER')

    def test_member_update_lbaasdriver(self):
        """Implements test case for update member method of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('UPDATE_MEMBER')

    def test_pool_health_monitor_create_lbaasdriver(self):
        """Implements test case for create pool_health_monitor method of
        loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('CREATE_POOL_HEALTH_MONITOR')

    def test_pool_health_monitor_delete_lbaasdriver(self):
        """Implements test case for delete pool_health_monitor method
        of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('DELETE_POOL_HEALTH_MONITOR')

    def test_pool_health_monitor_update_lbaasdriver(self):
        """Implements test case for update pool_health_monitor method
        of loadbalancer driver.

        Returns: none

        """

        self._test_lbaasdriver('UPDATE_POOL_HEALTH_MONITOR')



# Copyright (c) 2016 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions

from gbpservice.neutron.plugins.ml2plus import extension_overrides

# Monkeypatch Neutron to allow overriding its own extension
# descriptors. Note that extension descriptor classes cannot be
# monkeypatched directly because they are loaded explicitly by file
# name and then used immediately.

_real_get_extensions_path = extensions.get_extensions_path


def get_extensions_path(service_plugins=None):
    path = _real_get_extensions_path(service_plugins)
    return extension_overrides.__path__[0] + ':' + path

extensions.get_extensions_path = get_extensions_path


from neutron.db import models_v2
from neutron.plugins.ml2 import db as ml2_db
from neutron.plugins.ml2 import models
from oslo_log import log as logging
from sqlalchemy.orm import exc


LOG = logging.getLogger(__name__)


# REVISIT: This method gets decorated in Pike for removal in Queens. So this
# patching might need to be changed in Pike and removed in Queens.
def patched_get_locked_port_and_binding(session, port_id):
    """Get port and port binding records for update within transaction."""
    LOG.debug("Using patched_get_locked_port_and_binding")
    try:
        port = (session.query(models_v2.Port).
                enable_eagerloads(False).
                filter_by(id=port_id).
                one())
        binding = (session.query(models.PortBinding).
                   enable_eagerloads(False).
                   filter_by(port_id=port_id).
                   one())
        return port, binding
    except exc.NoResultFound:
        return None, None


ml2_db.get_locked_port_and_binding = patched_get_locked_port_and_binding



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

FIREWALL = 'firewall'
LOADBALANCER = 'loadbalancer'
LOADBALANCERV2 = 'loadbalancerv2'
VPN = 'vpn'
GENERIC_CONFIG = 'generic_config'

GBP_MODE = "gbp"
NEUTRON_MODE = "neutron"
NOVA_MODE = "nova"

NEUTRON_PORT = "neutron_port"
GBP_PORT = "gbp_policy_target"

NEUTRON_NETWORK = "neutron_network"
GBP_NETWORK = "gbp_group"

PROVIDER = "provider"
CONSUMER = "consumer"
STITCHING = "stitching"
MANAGEMENT = "management"
MONITOR = "monitoring"
GATEWAY_TYPE = "gateway"
ENDPOINT_TYPE = "endpoint"

CREATE = "create"
UPDATE = "update"
DELETE = "delete"

SUCCESS = 'SUCCESS'

FOREVER = 'forever'
INITIAL = 'initial'

ACTIVE_PORT = "ACTIVE"
STANDBY_PORT = "STANDBY"
MASTER_PORT = "MASTER"
STANDALONE_PORT = "STANDALONE"

ACTIVE = "ACTIVE"
# REVISIT(ashu) - Merge to have single BUILD state
PENDING_CREATE = "PENDING_CREATE"
PENDING_UPDATE = "PENDING_UPDATE"
PENDING_DELETE = "PENDING_DELETE"
ERROR = "ERROR"

DEVICE_ORCHESTRATOR = "device_orch"
SERVICE_ORCHESTRATOR = "service_orch"

HEAT_CONFIG_TAG = 'heat_config'
CONFIG_INIT_TAG = 'config_init'
ANSIBLE_TAG = 'ansible'
CUSTOM_JSON = 'custom_json'

COMPLETED = "COMPLETED"
IN_PROGRESS = "IN_PROGRESS"

CONFIG_INIT_TAG = "config_init"
CONFIG_SCRIPT = 'config_script'

CONFIG_TAG_RESOURCE_MAP = {
    HEAT_CONFIG_TAG: 'heat',
    CONFIG_INIT_TAG: 'config_init',
    ANSIBLE_TAG: 'ansible',
    CUSTOM_JSON: 'custom_json'}

MAXIMUM_INTERFACES = 'maximum_interfaces'
SUPPORTS_SHARING = 'supports_device_sharing'
SUPPORTS_HOTPLUG = 'supports_hotplug'

PERIODIC_HM = 'periodic_healthmonitor'
DEVICE_TO_BECOME_DOWN = 'DEVICE_TO_BECOME_DOWN'

METADATA_SUPPORTED_ATTRIBUTES = [MAXIMUM_INTERFACES,
                                 SUPPORTS_SHARING,
                                 SUPPORTS_HOTPLUG]

LOADBALANCER_RPC_API_VERSION = "2.0"
LOADBALANCERV2_RPC_API_VERSION = "1.0"

HEALTHMONITOR_RESOURCE = 'healthmonitor'
INTERFACE_RESOURCE = 'interfaces'
ROUTES_RESOURCE = 'routes'
MANAGEMENT_INTERFACE_NAME = 'mgmt_interface'

VYOS_VENDOR = 'vyos'
HAPROXY_VENDOR = 'haproxy'
HAPROXY_LBAASV2 = 'haproxy_lbaasv2'
NFP_VENDOR = 'nfp'
L3_INSERTION_MODE = "l3"

request_event = "REQUEST"
response_event = "RESPONSE"
error_event = "ERROR"

#POLLING EVENTS SPACING AND MAXRETRIES
DEVICE_SPAWNING_SPACING = 10
DEVICE_SPAWNING_MAXRETRY = 25

DEVICE_BEING_DELETED_SPACING = 5
DEVICE_BEING_DELETED_MAXRETRY = 20

APPLY_USER_CONFIG_IN_PROGRESS_SPACING = 10
APPLY_USER_CONFIG_IN_PROGRESS_MAXRETRY = 20

UPDATE_USER_CONFIG_PREPARING_TO_START_SPACING = 10
UPDATE_USER_CONFIG_PREPARING_TO_START_MAXRETRY = 40

UPDATE_USER_CONFIG_STILL_IN_PROGRESS_MAXRETRY = 300

DELETE_USER_CONFIG_IN_PROGRESS_SPACING = 10
DELETE_USER_CONFIG_IN_PROGRESS_MAXRETRY = 20

CHECK_USER_CONFIG_COMPLETE_SPACING = 10
CHECK_USER_CONFIG_COMPLETE_MAXRETRY = 40

PULL_NOTIFICATIONS_SPACING = 10

#nfp_node_deriver_config
# all units in sec.
SERVICE_CREATE_TIMEOUT = 1500
SERVICE_DELETE_TIMEOUT = 600

# heat stack creation timeout
STACK_ACTION_WAIT_TIME = 600


# default directory for config files
CONFIG_DIR = '/etc/nfp/'



# Copyright (c) 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

# Register apic_system_id
# REVISIT(ivar): would be nice to remove dependency from apic_ml2 in GBP, and
# register option directly here.
from apic_ml2.neutron.plugins.ml2.drivers.cisco.apic import config  # noqa


apic_opts = [
    cfg.BoolOpt('enable_optimized_dhcp', default=True),
    cfg.BoolOpt('enable_optimized_metadata', default=False),
    cfg.StrOpt('keystone_notification_exchange',
               default='keystone',
               help=("The exchange used to subscribe to Keystone "
                     "notifications")),
    cfg.StrOpt('keystone_notification_topic',
               default='notifications',
               help=("The topic used to subscribe to Keystone "
                     "notifications")),
    cfg.IntOpt('apic_optimized_dhcp_lease_time', default=0,
               help=("Number of seconds for the optimized DHCP lease time. "
                     "Default is 0 which means using opflex agent's default "
                     "value.")),
    cfg.BoolOpt('enable_keystone_notification_purge',
                default=False,
                help=("This will enable purging all the resources including "
                      "the tenant once a keystone project.deleted "
                      "notification is received.")),
]


cfg.CONF.register_opts(apic_opts, "ml2_apic_aim")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from gbpservice.contrib.nfp.configurator.agents import agent_base
from gbpservice.contrib.nfp.configurator.lib import data_filter
from gbpservice.contrib.nfp.configurator.lib import lb_constants
from gbpservice.contrib.nfp.configurator.lib import utils
from gbpservice.nfp.core import event as nfp_event
from gbpservice.nfp.core import log as nfp_logging
from gbpservice.nfp.core import module as nfp_api

from neutron._i18n import _LI
from neutron import context

LOG = nfp_logging.getLogger(__name__)


class LBaasRpcSender(data_filter.Filter):
    """Implements LBaaS response path to Neutron plugin.

    Methods of this class are invoked by LBaasEventHandler class
    and also by driver class for sending response from driver to
    the LBaaS Neutron plugin.

    """

    def __init__(self, sc):
        self.notify = agent_base.AgentBaseNotification(sc)

    def get_logical_device(self, pool_id, context):
        """ Calls data filter library to get logical device from pool_id.

        :param pool_id: Neutron LBaaS pool id
        :param context: RPC context

        Returns: logical_device

        """
        LOG.info(_LI("Sending RPC call 'Get Logical Device' "
                     "for pool: %(pool_id)s"),
                 {'pool_id': pool_id})
        return self.call(
            context,
            self.make_msg(
                'get_logical_device',
                pool_id=pool_id
            )
        )

    def update_status(self, obj_type, obj_id, status, agent_info, obj=None):
        """ Enqueues the response from LBaaS operation to neutron plugin.

        :param obj_type: Neutron LBaaS object type
        :param obj_id: Neutron LBaaS object id
        :param status: Neutron LBaaS object status to be set
        :param agent_info: Agent info which carries context which is needed
                           in config_orch to send response back to *aaS plugin
        :param obj: Neutron LBaaS object

        Returns: None

        """

        msg = {'info': {'service_type': lb_constants.SERVICE_TYPE,
                        'context': agent_info['context']},
               'notification': [{'resource': agent_info['resource'],
                                 'data': {'obj_type': obj_type,
                                          'obj_id': obj_id,
                                          'notification_type': 'update_status',
                                          'status': status,
                                          obj_type: obj}}]
               }
        LOG.info(_LI("Sending Notification 'Update Status' "
                     "for resource: %(resource)s with status:"
                     "%(status)s"),
                 {'resource': agent_info['resource'],
                  'status': status})
        self.notify._notification(msg)

    def update_pool_stats(self, pool_id, stats, context, pool=None):
        """ Enqueues the response from LBaaS operation to neutron plugin.

        :param pool_id: Neutron LBaaS pool id
        :param stats: statistics of that pool
        :param context: RPC context

        Returns: None
        """
        msg = {'info': {'service_type': lb_constants.SERVICE_TYPE,
                        'context': context.to_dict()},
               'notification': [{'resource': 'pool',
                                 'data': {'pool_id': pool_id,
                                          'stats': stats,
                                          'notification_type': (
                                              'update_pool_stats'),
                                          'pool': pool_id}}]
               }
        LOG.info(_LI("Sending Notification 'Update Pool Stats' "
                     "for pool: %(pool_id)s with stats:%(stats)s"),
                 {'pool_id': pool_id,
                  'stats': stats})
        self.notify._notification(msg)

    def vip_deleted(self, vip, status, agent_info):
        """ Enqueues the response from LBaaS operation to neutron plugin.

        :param vip: Neutron LBaaS vip resource
        :param status: Neutron LBaaS vip resource status
        :param agent_info: Agent info which carries context which is needed
                           in config_orch to send response back to *aaS plugin

        Returns: None
        """
        msg = {'info': {'service_type': lb_constants.SERVICE_TYPE,
                        'context': agent_info['context']},
               'notification': [{'resource': agent_info['resource'],
                                 'data': {'vip_id': vip['id'],
                                          'vip': vip,
                                          'notification_type': 'vip_deleted',
                                          'status': status}}]
               }
        LOG.info(_LI("Sending Notification 'VIP Deleted' "
                     "for vip: %(vip_id)s with status:%(status)s"),
                 {'vip_id': vip['id'],
                  'status': status})
        self.notify._notification(msg)


class LBaaSRpcManager(agent_base.AgentBaseRPCManager):
    """Implements APIs invoked by configurator for processing RPC messages.

    RPC client of configurator module receives RPC messages from REST server
    and invokes the API of this class. The instance of this class is registered
    with configurator module using register_service_agent API. Configurator
    module identifies the service agent object based on service type and
    invokes one of the methods of this class to configure the device.

    """

    def __init__(self, sc, conf):
        """Instantiates child and parent class objects.

        :param sc: Service Controller object that is used for interfacing
        with core service controller.
        :param conf: Configuration object that is used for configuration
        parameter access.

        """

        super(LBaaSRpcManager, self).__init__(sc, conf)

    def _send_event(self, event_id, data, serialize=False, binding_key=None,
                    key=None):
        """Posts an event to framework.

        :param event_id: Unique identifier for the event
        :param data: event data
        :param serialize: boolean value used to serialize the event
        :param binding_key: binding key to be used for serialization
        :param key: event key

        Returns: None

        """

        ev = self.sc.new_event(id=event_id, data=data)
        ev.key = key
        ev.sequence = serialize
        ev.binding_key = binding_key
        self.sc.post_event(ev)

    def create_vip(self, context, vip):
        """Enqueues event for worker to process create vip request.

        :param context: RPC context
        :param vip: vip resource to be created

        Returns: None

        """
        LOG.info(_LI("Received request 'Create VIP' for VIP:%(vip_id)s "),
                 {'vip_id': vip['id']})
        arg_dict = {'context': context,
                    'vip': vip,
                    }
        self._send_event(lb_constants.EVENT_CREATE_VIP, arg_dict,
                         serialize=True, binding_key=vip['pool_id'],
                         key=vip['id'])

    def update_vip(self, context, old_vip, vip):
        """Enqueues event for worker to process update vip request.

        :param context: RPC context
        :param old_vip: old vip resource to be updated
        :param vip: new vip resource

        Returns: None

        """
        old_val, new_val = self.get_diff_of_dict(old_vip, vip)
        arg_dict = {'context': context,
                    'old_vip': old_vip,
                    'vip': vip,
                    }
        LOG.info(_LI("Received request 'Update VIP' for VIP:%(vip_id)s "
                     "with new Param:%(new_val)s and old Param:%(old_val)s"),
                 {'vip_id': vip['id'],
                  'new_val': new_val,
                  'old_val': old_val})
        self._send_event(lb_constants.EVENT_UPDATE_VIP, arg_dict,
                         serialize=True, binding_key=vip['pool_id'],
                         key=vip['id'])

    def delete_vip(self, context, vip):
        """Enqueues event for worker to process delete vip request.

        :param context: RPC context
        :param vip: vip resource to be deleted

        Returns: None

        """
        arg_dict = {'context': context,
                    'vip': vip,
                    }
        LOG.info(_LI("Received 'Delete VIP' request for VIP:%(vip_id)s "),
                 {'vip_id': vip['id']})
        self._send_event(lb_constants.EVENT_DELETE_VIP, arg_dict,
                         serialize=True, binding_key=vip['pool_id'],
                         key=vip['id'])

    def create_pool(self, context, pool, driver_name):
        """Enqueues event for worker to process create pool request.

        :param context: RPC context
        :param pool: pool resource to be created
        :param driver_name: service vendor driver name

        Returns: None

        """
        arg_dict = {'context': context,
                    'pool': pool,
                    'driver_name': driver_name,
                    }
        LOG.info(_LI("Received request 'Create Pool' for Pool:%(pool_id)s "),
                 {'pool_id': pool['id']})
        self._send_event(lb_constants.EVENT_CREATE_POOL, arg_dict,
                         serialize=True, binding_key=pool['id'],
                         key=pool['id'])

    def update_pool(self, context, old_pool, pool):
        """Enqueues event for worker to process update pool request.

        :param context: RPC context
        :param old_pool: old pool resource to be updated
        :param pool: new pool resource

        Returns: None

        """
        old_val, new_val = self.get_diff_of_dict(old_pool, pool)
        arg_dict = {'context': context,
                    'old_pool': old_pool,
                    'pool': pool,
                    }
        LOG.info(_LI("Received request 'Update Pool' for Pool:%(pool_id)s "
                     "with new Param:%(new_val)s and old Param:%(old_val)s"),
                 {'pool_id': pool['id'],
                  'new_val': new_val,
                  'old_val': old_val})
        self._send_event(lb_constants.EVENT_UPDATE_POOL, arg_dict,
                         serialize=True, binding_key=pool['id'],
                         key=pool['id'])

    def delete_pool(self, context, pool):
        """Enqueues event for worker to process delete pool request.

        :param context: RPC context
        :param pool: pool resource to be deleted

        Returns: None

        """
        arg_dict = {'context': context,
                    'pool': pool,
                    }
        LOG.info(_LI("Received request 'Delete Pool' for Pool:%(pool_id)s "),
                 {'pool_id': pool['id']})
        self._send_event(lb_constants.EVENT_DELETE_POOL, arg_dict,
                         serialize=True, binding_key=pool['id'],
                         key=pool['id'])

    def create_member(self, context, member):
        """Enqueues event for worker to process create member request.

        :param context: RPC context
        :param member: member resource to be created

        Returns: None

        """
        arg_dict = {'context': context,
                    'member': member,
                    }
        LOG.info(_LI("Received request 'Create Member' for Pool:"
                     "%(pool_id)s "),
                 {'pool_id': member['pool_id']})
        self._send_event(lb_constants.EVENT_CREATE_MEMBER, arg_dict,
                         serialize=True, binding_key=member['pool_id'],
                         key=member['id'])

    def update_member(self, context, old_member, member):
        """Enqueues event for worker to process update member request.

        :param context: RPC context
        :param old_member: old member resource to be updated
        :param member: new member resource

        Returns: None

        """
        old_val, new_val = self.get_diff_of_dict(old_member, member)
        arg_dict = {'context': context,
                    'old_member': old_member,
                    'member': member,
                    }
        LOG.info(_LI("Received request 'Update Member' for Member:"
                     "%(member_id)s in Pool:%(pool_id)s with new Param:"
                     "%(new_val)s and old Param:%(old_val)s"),
                 {'member_id': member['id'],
                  'pool_id': member['pool_id'],
                  'new_val': new_val,
                  'old_val': old_val})
        self._send_event(lb_constants.EVENT_UPDATE_MEMBER, arg_dict,
                         serialize=True, binding_key=member['pool_id'],
                         key=member['id'])

    def delete_member(self, context, member):
        """Enqueues event for worker to process delete member request.

        :param context: RPC context
        :param member: member resource to be deleted

        Returns: None

        """
        arg_dict = {'context': context,
                    'member': member,
                    }
        LOG.info(_LI("Received request 'Delete Member' for Pool:%(pool_id)s "),
                 {'pool_id': member['pool_id']})
        self._send_event(lb_constants.EVENT_DELETE_MEMBER, arg_dict,
                         serialize=True, binding_key=member['pool_id'],
                         key=member['id'])

    def create_pool_health_monitor(self, context, health_monitor, pool_id):
        """Enqueues event for worker to process create health monitor request.

        :param context: RPC context
        :param health_monitor: health_monitor resource to be created
        :param pool_id: pool_id to which health monitor is associated

        Returns: None

        """
        arg_dict = {'context': context,
                    'health_monitor': health_monitor,
                    'pool_id': pool_id,
                    }
        LOG.info(_LI("Received request 'Create Pool Health Monitor' for "
                     "Pool:%(pool_id)s and Health monitor:%(hm)s"),
                 {'pool_id': pool_id,
                  'hm': health_monitor['id']})
        self._send_event(lb_constants.EVENT_CREATE_POOL_HEALTH_MONITOR,
                         arg_dict, serialize=True, binding_key=pool_id,
                         key=health_monitor['id'])

    def update_pool_health_monitor(self, context, old_health_monitor,
                                   health_monitor, pool_id):
        """Enqueues event for worker to process update health monitor request.

        :param context: RPC context
        :param old_health_monitor: health_monitor resource to be updated
        :param health_monitor: new health_monitor resource
        :param pool_id: pool_id to which health monitor is associated

        Returns: None

        """
        old_val, new_val = self.get_diff_of_dict(
            old_health_monitor, health_monitor)
        arg_dict = {'context': context,
                    'old_health_monitor': old_health_monitor,
                    'health_monitor': health_monitor,
                    'pool_id': pool_id,
                    }
        LOG.info(_LI("Received request 'Update Pool Health Monitor' for "
                     "Pool Health Monitor:%(hm_id)s with new Param:"
                     "%(new_val)s and old Param:%(old_val)s"),
                 {'hm_id': health_monitor['id'],
                  'new_val': new_val,
                  'old_val': old_val})
        self._send_event(lb_constants.EVENT_UPDATE_POOL_HEALTH_MONITOR,
                         arg_dict, serialize=True, binding_key=pool_id,
                         key=health_monitor['id'])

    def delete_pool_health_monitor(self, context, health_monitor, pool_id):
        """Enqueues event for worker to process delete health monitor request.

        :param context: RPC context
        :param health_monitor: health_monitor resource to be deleted
        :param pool_id: pool_id to which health monitor is associated

        Returns: None

        """
        arg_dict = {'context': context,
                    'health_monitor': health_monitor,
                    'pool_id': pool_id,
                    }
        LOG.info(_LI("Received request 'Delete Pool Health Monitor' for "
                     "Pool:%(pool_id)s and Health monitor:%(hm)s"),
                 {'pool_id': pool_id,
                  'hm': health_monitor['id']})
        self._send_event(lb_constants.EVENT_DELETE_POOL_HEALTH_MONITOR,
                         arg_dict, serialize=True, binding_key=pool_id,
                         key=health_monitor['id'])

    def agent_updated(self, context, payload):
        """Enqueues event for worker to process agent updated request.

        :param context: RPC context
        :param payload: payload

        Returns: None

        """
        arg_dict = {'context': context,
                    'payload': payload}
        LOG.info(_LI("Received request 'Agent Updated' "))
        self._send_event(lb_constants.EVENT_AGENT_UPDATED, arg_dict)


class LBaaSEventHandler(agent_base.AgentBaseEventHandler,
                        nfp_api.NfpEventHandler):
    """Implements event handlers and their helper methods.

    Object of this class is registered with the event class of core service
    controller. Based on the event key, handle_event method of this class is
    invoked by core service controller.

    """
    instance_mapping = {}

    def __init__(self, sc, drivers, rpcmgr):
        self.sc = sc
        self.drivers = drivers
        self.rpcmgr = rpcmgr
        self.plugin_rpc = LBaasRpcSender(sc)

        """REVISIT (pritam):
        Remove neutron context dependency. As of now
        because config agent needs context in notification, and internal
        poll event like collect_stats() does not have context, creating
        context here, but should get rid of this in future.
        """
        self.context = context.get_admin_context_without_session()

    def _get_driver(self, service_vendor, service_feature):
        """Retrieves service driver instance based on service type
        and service vendor.

        :param service_vendor: service vendor

        Returns: Service driver instance

        """
        driver = lb_constants.SERVICE_TYPE + service_vendor + service_feature
        return self.drivers[driver]

    def handle_event(self, ev):
        """Processes generated events in worker context.

        Processes following events.
        - create pool
        - update pool
        - delete pool
        - create vip
        - update vip
        - delete vip
        - create member
        - update member
        - delete member
        - create pool health monitor
        - update pool health monitor
        - delete pool health monitor
        - agent updated
        Enqueues responses into notification queue.

        Returns: None

        """
        msg = ("Starting handling event '%s' " % (ev.id))
        LOG.info(msg)
        try:
            method = getattr(self, "_%s" % (ev.id.lower()))
            method(ev)
        except Exception as err:
            msg = ("Failed to handle event %s. Reason is %s"
                   % (ev.id, str(err).capitalize()))
            LOG.error(msg)
        finally:
            if ev.id == lb_constants.EVENT_COLLECT_STATS:
                """Do not say event done for this event as it is
                   to be executed forever
                """
                pass
            else:
                msg = ("Completed handling event '%s'" % (ev.id))
                LOG.info(msg)
                self.sc.event_complete(ev)

    def _handle_event_vip(self, ev, operation):
        data = ev.data
        context = data['context']
        vip = data['vip']
        agent_info = ev.data['context'].pop('agent_info')
        service_vendor = agent_info['service_vendor']
        service_feature = agent_info['service_feature']
        driver = self._get_driver(service_vendor, service_feature)

        try:
            if operation == lb_constants.CREATE:
                driver.create_vip(vip, context)
            elif operation == lb_constants.UPDATE:
                old_vip = data['old_vip']
                driver.update_vip(old_vip, vip, context)
            elif operation == lb_constants.DELETE:
                driver.delete_vip(vip, context)
                self.plugin_rpc.vip_deleted(vip,
                                            lb_constants.ACTIVE, agent_info)
                return  # Don't update object status for delete operation
        except Exception as e:
            if operation == lb_constants.DELETE:
                msg = ("Failed to delete vip %s" % (vip['id']))
                self.plugin_rpc.vip_deleted(vip,
                                            lb_constants.ACTIVE, agent_info)
                LOG.warn(msg)
            else:
                self.plugin_rpc.update_status('vip', vip['id'],
                                              lb_constants.ERROR,
                                              agent_info, vip)
                raise e
        else:
            self.plugin_rpc.update_status('vip', vip['id'],
                                          lb_constants.ACTIVE,
                                          agent_info, vip)

    def _create_vip(self, ev):
        self._handle_event_vip(ev, lb_constants.CREATE)

    def _update_vip(self, ev):
        self._handle_event_vip(ev, lb_constants.UPDATE)

    def _delete_vip(self, ev):
        self._handle_event_vip(ev, lb_constants.DELETE)

    def _handle_event_pool(self, ev, operation):
        data = ev.data
        context = data['context']
        pool = data['pool']
        agent_info = context.pop('agent_info')
        service_vendor = agent_info['service_vendor']
        service_feature = agent_info['service_feature']
        try:
            if operation == lb_constants.CREATE:
                driver_name = data['driver_name']
                driver_id = driver_name + service_vendor + service_feature
                if (driver_id) not in self.drivers.keys():
                    msg = ('No device driver on agent: %s.' % (driver_name))
                    LOG.error(msg)
                    self.plugin_rpc.update_status('pool', pool['id'],
                                                  lb_constants.ERROR,
                                                  agent_info, pool)
                    return
                driver = self.drivers[driver_id]
                driver.create_pool(pool, context)
                LBaaSEventHandler.instance_mapping[pool['id']] = driver_name
            elif operation == lb_constants.UPDATE:
                old_pool = data['old_pool']
                driver = self._get_driver(service_vendor,
                                          service_feature)  # pool['id'])
                driver.update_pool(old_pool, pool, context)
            elif operation == lb_constants.DELETE:
                driver = self._get_driver(service_vendor,
                                          service_feature)  # pool['id'])
                driver.delete_pool(pool, context)
                del LBaaSEventHandler.instance_mapping[pool['id']]
                return  # Don't update object status for delete operation
        except Exception as e:
            if operation == lb_constants.DELETE:
                msg = ("Failed to delete pool %s" % (pool['id']))
                LOG.warn(msg)
                del LBaaSEventHandler.instance_mapping[pool['id']]
            else:
                self.plugin_rpc.update_status('pool', pool['id'],
                                              lb_constants.ERROR,
                                              agent_info, pool)
                raise e
        else:
            self.plugin_rpc.update_status('pool', pool['id'],
                                          lb_constants.ACTIVE,
                                          agent_info, pool)

    def _create_pool(self, ev):
        self._handle_event_pool(ev, lb_constants.CREATE)

    def _update_pool(self, ev):
        self._handle_event_pool(ev, lb_constants.UPDATE)

    def _delete_pool(self, ev):
        self._handle_event_pool(ev, lb_constants.DELETE)

    def _handle_event_member(self, ev, operation):
        data = ev.data
        context = data['context']
        member = data['member']
        agent_info = ev.data['context'].pop('agent_info')
        service_vendor = agent_info['service_vendor']
        service_feature = agent_info['service_feature']
        driver = self._get_driver(service_vendor,
                                  service_feature)  # member['pool_id'])
        try:
            if operation == lb_constants.CREATE:
                driver.create_member(member, context)
            elif operation == lb_constants.UPDATE:
                old_member = data['old_member']
                driver.update_member(old_member, member, context)
            elif operation == lb_constants.DELETE:
                driver.delete_member(member, context)
                return  # Don't update object status for delete operation
        except Exception as e:
            if operation == lb_constants.DELETE:
                msg = ("Failed to delete member %s" % (member['id']))
                LOG.warn(msg)
            else:
                self.plugin_rpc.update_status('member', member['id'],
                                              lb_constants.ERROR,
                                              agent_info, member)
                raise e
        else:
            self.plugin_rpc.update_status('member', member['id'],
                                          lb_constants.ACTIVE,
                                          agent_info, member)

    def _create_member(self, ev):
        self._handle_event_member(ev, lb_constants.CREATE)

    def _update_member(self, ev):
        self._handle_event_member(ev, lb_constants.UPDATE)

    def _delete_member(self, ev):
        self._handle_event_member(ev, lb_constants.DELETE)

    def _handle_event_pool_health_monitor(self, ev, operation):
        data = ev.data
        context = data['context']
        agent_info = context.pop('agent_info')
        health_monitor = data['health_monitor']
        pool_id = data['pool_id']
        service_vendor = agent_info['service_vendor']
        service_feature = agent_info['service_feature']
        driver = self._get_driver(service_vendor, service_feature)  # (pool_id)
        assoc_id = {'pool_id': pool_id,
                    'monitor_id': health_monitor['id']}
        try:
            if operation == lb_constants.CREATE:
                driver.create_pool_health_monitor(health_monitor, pool_id,
                                                  context)
            elif operation == lb_constants.UPDATE:
                old_health_monitor = data['old_health_monitor']
                driver.update_pool_health_monitor(old_health_monitor,
                                                  health_monitor, pool_id,
                                                  context)
            elif operation == lb_constants.DELETE:
                driver.delete_pool_health_monitor(health_monitor, pool_id,
                                                  context)
                return  # Don't update object status for delete operation
        except Exception as e:
            if operation == lb_constants.DELETE:
                msg = ("Failed to delete pool health monitor."
                       " assoc_id: %s" % (assoc_id))
                LOG.warn(msg)
            else:
                self.plugin_rpc.update_status(
                    'health_monitor', assoc_id, lb_constants.ERROR,
                    agent_info, health_monitor)
                raise e
        else:
            self.plugin_rpc.update_status(
                'health_monitor', assoc_id, lb_constants.ACTIVE,
                agent_info, health_monitor)

    def _create_pool_health_monitor(self, ev):
        self._handle_event_pool_health_monitor(ev, lb_constants.CREATE)

    def _update_pool_health_monitor(self, ev):
        self._handle_event_pool_health_monitor(ev, lb_constants.UPDATE)

    def _delete_pool_health_monitor(self, ev):
        self._handle_event_pool_health_monitor(ev, lb_constants.DELETE)

    def _agent_updated(self, ev):
        """ REVISIT(pritam): Implement this method """
        return None

    def _collect_stats(self, ev):
        self.sc.poll_event(ev)

    @nfp_api.poll_event_desc(event=lb_constants.EVENT_COLLECT_STATS,
                             spacing=60)
    def collect_stats(self, ev):
        for pool_id, driver_name in LBaaSEventHandler.instance_mapping.items():
            driver_id = lb_constants.SERVICE_TYPE + driver_name
            driver = self.drivers[driver_id]
            try:
                stats = driver.get_stats(pool_id)
                if stats:
                    self.plugin_rpc.update_pool_stats(pool_id, stats,
                                                      self.context)
            except Exception:
                msg = ("Error updating statistics on pool %s" % (pool_id))
                LOG.error(msg)


def events_init(sc, drivers, rpcmgr):
    """Registers events with core service controller.

    All the events will come to handle_event method of class instance
    registered in 'handler' field.

    :param drivers: Driver instances registered with the service agent
    :param rpcmgr: Instance to receive all the RPC messages from configurator
    module.

    Returns: None

    """
    ev_ids = [lb_constants.EVENT_CREATE_POOL, lb_constants.EVENT_UPDATE_POOL,
              lb_constants.EVENT_DELETE_POOL,

              lb_constants.EVENT_CREATE_VIP, lb_constants.EVENT_UPDATE_VIP,
              lb_constants.EVENT_DELETE_VIP,

              lb_constants.EVENT_CREATE_MEMBER,
              lb_constants.EVENT_UPDATE_MEMBER,
              lb_constants.EVENT_DELETE_MEMBER,

              lb_constants.EVENT_CREATE_POOL_HEALTH_MONITOR,
              lb_constants.EVENT_UPDATE_POOL_HEALTH_MONITOR,
              lb_constants.EVENT_DELETE_POOL_HEALTH_MONITOR,

              lb_constants.EVENT_AGENT_UPDATED,
              lb_constants.EVENT_COLLECT_STATS
              ]

    evs = []
    for ev_id in ev_ids:
        ev = nfp_event.Event(id=ev_id, handler=LBaaSEventHandler(sc, drivers,
                                                                 rpcmgr))
        evs.append(ev)
    sc.register_events(evs)


def load_drivers(sc, conf):
    """Imports all the driver files.

    Returns: Dictionary of driver objects with a specified service type and/or
    vendor name

    """
    cutils = utils.ConfiguratorUtils(conf)
    drivers = cutils.load_drivers(lb_constants.SERVICE_TYPE)

    plugin_rpc = LBaasRpcSender(sc)

    for service_type, dobj in drivers.iteritems():
        '''LB Driver constructor needs plugin_rpc as a param'''
        instantiated_dobj = dobj(plugin_rpc=plugin_rpc, conf=conf)
        drivers[service_type] = instantiated_dobj

    return drivers


def register_service_agent(cm, sc, conf, rpcmgr):
    """Registers Loadbalaner service agent with configurator module.

    :param cm: Instance of configurator module
    :param sc: Instance of core service controller
    :param conf: Instance of oslo configuration
    :param rpcmgr: Instance containing RPC methods which are invoked by
    configurator module on corresponding RPC message arrival

    """

    service_type = lb_constants.SERVICE_TYPE
    cm.register_service_agent(service_type, rpcmgr)


def init_agent(cm, sc, conf):
    """Initializes Loadbalaner agent.

    :param cm: Instance of configuration module
    :param sc: Instance of core service controller
    :param conf: Instance of oslo configuration

    """

    try:
        drivers = load_drivers(sc, conf)
    except Exception as err:
        msg = ("Loadbalaner agent failed to load service drivers. Reason:%s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = ("Loadbalaner agent loaded service"
               " drivers successfully.")
        LOG.debug(msg)

    rpcmgr = LBaaSRpcManager(sc, conf)

    try:
        events_init(sc, drivers, rpcmgr)
    except Exception as err:
        msg = ("Loadbalaner agent failed to initialize events. Reason:%s"
               % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = ("Loadbalaner agent initialized"
               " events successfully.")
        LOG.debug(msg)

    try:
        register_service_agent(cm, sc, conf, rpcmgr)
    except Exception as err:
        msg = ("Failed to register Loadbalaner agent with"
               " configurator module. Reason:%s" % (str(err).capitalize()))
        LOG.error(msg)
        raise err
    else:
        msg = ("Loadbalaner agent registered with configuration"
               " module successfully.")
        LOG.debug(msg)


def _start_collect_stats(sc):
    """Enqueues poll event for worker to collect pool stats periodically.
       Agent keeps map of pool_id:driver. As part of handling this event,
       stats for pool_id is requested from agent inside service vm
    """

    arg_dict = {}
    ev = sc.new_event(id=lb_constants.EVENT_COLLECT_STATS, data=arg_dict)
    sc.post_event(ev)


def init_agent_complete(cm, sc, conf):
    # _start_collect_stats(sc)
    msg = ("Initialization of loadbalancer agent completed.")
    LOG.info(msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.common import config  # noqa
from neutron import context as n_context
from neutron import manager
from oslo_config import cfg

from gbpservice.neutron.services.servicechain.plugins.ncp import model
from gbpservice.neutron.tests.unit.services.grouppolicy import (
    test_resource_mapping as test_gp_driver)
from gbpservice.neutron.tests.unit.services.servicechain.ncp import (
    test_ncp_plugin as base)


GATEWAY = 'gateway'
GATEWAY_HA = 'gateway_ha'
TRANSPARENT = 'transparent'
ENDPOINT = 'endpoint'

info_mapping = {
    GATEWAY: {'plumbing_type': GATEWAY, 'provider': [{}], 'consumer': [{}]},
    GATEWAY_HA: {'plumbing_type': GATEWAY, 'provider': [{}, {}, {}],
                 'consumer': [{}, {}, {}]},
    TRANSPARENT: {'plumbing_type': TRANSPARENT, 'provider': [{}],
                  'consumer': [{}]},
    ENDPOINT: {'plumbing_type': ENDPOINT, 'provider': [{}]},

}
info_mapping['FIREWALL'] = info_mapping[GATEWAY]
info_mapping['FIREWALL_HA'] = info_mapping[GATEWAY_HA]
info_mapping['TRANSPARENT_FIREWALL'] = info_mapping[TRANSPARENT]
info_mapping['LOADBALANCER'] = info_mapping[ENDPOINT]


class ResourceMappingStitchingPlumberGBPTestCase(
        test_gp_driver.ResourceMappingTestCase):

    def setUp(self):
        cfg.CONF.set_override(
            'extension_drivers', ['proxy_group'], group='group_policy')
        cfg.CONF.set_override('node_plumber', 'stitching_plumber',
                              group='node_composition_plugin')
        ml2_opts = {'mechanism_drivers': ['stitching_gbp']}
        host_agents = mock.patch('neutron.plugins.ml2.driver_context.'
                                 'PortContext.host_agents').start()
        host_agents.return_value = [self.agent_conf]
        super(ResourceMappingStitchingPlumberGBPTestCase, self).setUp(
            sc_plugin=base.SC_PLUGIN_KLASS, ml2_options=ml2_opts)

        def get_plumbing_info(context):
            return info_mapping.get(context.current_profile['service_type'])

        self.node_driver = self.sc_plugin.driver_manager.ordered_drivers[0].obj
        self.node_driver.get_plumbing_info = get_plumbing_info

    @property
    def sc_plugin(self):
        plugins = manager.NeutronManager.get_service_plugins()
        servicechain_plugin = plugins.get('SERVICECHAIN')
        return servicechain_plugin


class TestPolicyRuleSet(ResourceMappingStitchingPlumberGBPTestCase,
                        test_gp_driver.TestPolicyRuleSet):
    pass


class TestServiceChain(ResourceMappingStitchingPlumberGBPTestCase,
                       test_gp_driver.TestServiceChain):

    def test_parent_ruleset_update_for_redirect(self):
        # NCP doesn't support multiple SPECs per instance
        pass

    def test_enforce_parent_redirect_after_ptg_create(self):
        # NCP doesn't support multiple SPECs per instance
        pass

    def test_hierarchical_redirect(self):
        # NCP doesn't support multiple SPECs per instance
        pass

    def test_redirect_multiple_ptgs_single_prs(self):
        # REVISIT(ivar): This test is doing a mock patching that breaks the
        # workflow
        pass

    def test_action_spec_value_update(self):
        # NCP doesn't support multiple SPECs per instance
        pass

    def test_rule_update_hierarchial_prs(self):
        # NCP doesn't support multiple SPECs per instance
        pass

    def test_rule_update_updates_chain(self):
        # NCP doesn't support multiple SPECs per instance
        pass


class TestServiceChainAdminOwner(ResourceMappingStitchingPlumberGBPTestCase,
                                 test_gp_driver.TestServiceChainAdminOwner):

    def test_parent_ruleset_update_for_redirect(self):
        # NCP doesn't support multiple SPECs per instance
        pass

    def test_enforce_parent_redirect_after_ptg_create(self):
        # NCP doesn't support multiple SPECs per instance
        pass

    def test_hierarchical_redirect(self):
        # NCP doesn't support multiple SPECs per instance
        pass

    def test_redirect_multiple_ptgs_single_prs(self):
        # REVISIT(ivar): This test is doing a mock patching that breaks the
        # workflow
        pass

    def test_action_spec_value_update(self):
        # NCP doesn't support multiple SPECs per instance
        pass

    def test_rule_update_hierarchial_prs(self):
        # NCP doesn't support multiple SPECs per instance
        pass

    def test_rule_update_updates_chain(self):
        # NCP doesn't support multiple SPECs per instance
        pass


class TestPolicyAction(ResourceMappingStitchingPlumberGBPTestCase,
                       test_gp_driver.TestPolicyAction):
    pass


class TestPolicyRule(ResourceMappingStitchingPlumberGBPTestCase,
                     test_gp_driver.TestPolicyRule):
    pass


class TestExternalSegment(ResourceMappingStitchingPlumberGBPTestCase,
                          test_gp_driver.TestExternalSegment):
    def test_update(self):
        super(TestExternalSegment, self).test_update(
            proxy_ip_pool1='182.169.0.0/16',
            proxy_ip_pool2='172.169.0.0/16')


class TestExternalPolicy(ResourceMappingStitchingPlumberGBPTestCase,
                         test_gp_driver.TestExternalPolicy):
    pass


class TestImplicitServiceChains(ResourceMappingStitchingPlumberGBPTestCase,
                                base.NodeCompositionPluginTestMixin):

    def test_service_targets_vif_details(self):
        context = n_context.get_admin_context()
        self._create_simple_service_chain(service_type='TRANSPARENT_FIREWALL')
        targets = model.get_service_targets(context.session)
        self.assertTrue(len(targets) > 0)
        for target in targets:
            pt = self.show_policy_target(
                target.policy_target_id)['policy_target']
            # Being service targets, port filter and hybrid plug will be false
            port = self._bind_port_to_host(pt['port_id'], 'host')['port']
            self.assertFalse(port['binding:vif_details']['port_filter'])
            self.assertFalse(port['binding:vif_details']['ovs_hybrid_plug'])

    def test_endpoint_target_vif_details(self):
        context = n_context.get_admin_context()
        self._create_simple_service_chain(service_type='LOADBALANCER')
        targets = model.get_service_targets(context.session)
        self.assertTrue(len(targets) > 0)
        for target in targets:
            pt = self.show_policy_target(
                target.policy_target_id)['policy_target']
            port = self._bind_port_to_host(pt['port_id'], 'host')['port']
            self.assertTrue(port['binding:vif_details']['port_filter'])
            # This change sets hybrid VIF plugging to True by default again
            # https://github.com/openstack/neutron/commit/
            # eca893be5b770c41cfc570dc016a41c30c2cdf23
            self.assertTrue(port['binding:vif_details']['ovs_hybrid_plug'])
